#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Linearity models."""

from collections.abc import Sequence

import numpy as np
from astropy import constants as const

from pyxel.detectors import CMOS, Detector
from pyxel.models.charge_measurement.non_linearity_calculation import (
    euler,
    hgcdte_bandgap,
    ni_hansen,
)


def compute_poly_linearity(
    array_2d: np.ndarray,
    coefficients: Sequence[float],
) -> np.ndarray:
    """Add non-linearity to an array of values following a polynomial function.

    Parameters
    ----------
    array_2d : ndarray
        Input array.
    coefficients : list of float
        Coefficients of the polynomial function.

    Returns
    -------
    np.ndarray
        Signal.
    """
    polynomial_function = np.polynomial.polynomial.Polynomial(coefficients)

    non_linear_signal = polynomial_function(array_2d)

    return non_linear_signal


def output_node_linearity_poly(
    detector: Detector,
    coefficients: Sequence[float],
) -> None:
    """Add non-linearity to signal array to simulate the non-linearity of the output node circuit.

    The non-linearity is simulated by a polynomial function. The user specifies the polynomial coefficients.

    detector Signal unit: Volt

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    coefficients : list of float
        Coefficient of the polynomial function.
    """
    if len(coefficients) == 0:
        raise ValueError("Length of coefficient list should be more than 0.")

    signal_mean_array = detector.signal.array.astype("float64")
    signal_non_linear = compute_poly_linearity(
        array_2d=signal_mean_array, coefficients=coefficients
    )

    if np.any(signal_non_linear < 0):
        raise ValueError(
            "Signal array contains negative values after applying non-linearity model!"
        )

    detector.signal.array = signal_non_linear


def compute_simple_physical_non_linearity(
    array_2d: np.ndarray,
    temperature: float,  # Detector operating temperature
    v_bias: float,
    cutoff: float,
    n_acceptor: float,
    n_donor: float,
    diode_diameter: float,
) -> np.ndarray:
    """Compute simple physical non-linear signal.

    Parameters
    ----------
    array_2d : ndarray
        Input array.
    temperature
        Temperature. Unit: K.
    v_bias : float
        Initial bias voltage. Unit: V.
    cutoff : float
        Cutoff wavelength. unit: um
    n_acceptor : float
        Acceptor density. Unit: atoms/cm^3
    n_donor : float
        Donor density. Unit: atoms/cm^3
    diode_diameter : float
        Diode diameter. Unit: um

    Returns
    -------
    ndarray
        Output array.
    """
    # Derivation of Cd concentration in the alloy,  it depends on cutoff wavelength and targeted operating temperature
    # Here we are considering the case where the detector is operated at its nominal temperature,
    # it might not be always the case
    # cutoff = 2.1
    e_g_targeted = 1.24 / cutoff  # cutoff is um and Eg in eV
    xcd = np.linspace(0.2, 0.6, 1000)
    targeted_operating_temperature = temperature
    e_g_calculated = hgcdte_bandgap(
        x_cd=xcd,
        temperature=targeted_operating_temperature,
    )  # Expected bandgap

    index = np.where(e_g_calculated > e_g_targeted)[0][0]
    x_cd = xcd[index]  # Targeted cadmium concentration in the HgCdTe alloy

    if not (0.2 <= x_cd <= 0.6):
        raise ValueError(
            "Hansen bangap expression used out of its nominal application range. "
            "x_cd must be between 0.2 and 0.6"
        )

    ni = ni_hansen(x_cd=x_cd, temperature=temperature)

    # Build in potential
    vbi = (
        const.k_B.value
        * temperature
        / const.e.value
        * np.log(n_acceptor * n_donor / ni**2)
    )

    # HgCdTe dielectric constant
    eps = 20.5 - 15.6 * x_cd + 5.7 * x_cd**2

    # Surface of the diode, assumed to be planar
    surface = (
        np.pi * (diode_diameter / 2.0 * 1e-6) ** 2
    )  # Surface of the diode, assumed to be circular

    # Initial value  of the diode capacitance
    co = surface * np.sqrt(
        (const.e.value * eps * const.eps0.value)
        / (2 * (vbi - v_bias))
        * ((n_acceptor * 1e6 * n_donor * 1e6) / (n_acceptor * 1e6 + n_donor * 1e6))
    )

    non_linear_signal = (
        1
        / (2 * vbi)
        * (const.e.value * array_2d / co) ** 2
        * (-1 + np.sqrt(1 + 4 * (co / (const.e.value * array_2d)) ** 2 * vbi**2))
    )

    return non_linear_signal


def simple_physical_non_linearity(
    detector: Detector,
    cutoff: float,
    n_acceptor: float,
    n_donor: float,
    diode_diameter: float,
    v_bias: float,
) -> None:
    """Apply simple physical non-linearity.

    Parameters
    ----------
    detector : Detector
        Pyxel detector object.
    cutoff : float
        Cutoff wavelength. unit: um
    n_donor : float
        Donor density. Unit: atoms/cm^3
    n_acceptor : float
        Acceptor density. Unit: atoms/cm^3
    diode_diameter : float
        Diode diameter. Unit: um
    v_bias : float
        Initial bias voltage. Unit: V.
    """

    if not (4 <= detector.environment.temperature <= 300):
        raise ValueError(
            "Hansen bangap expression used out of its nominal application range. "
            "temperature must be between 4K and 300K"
        )

    if not isinstance(detector, CMOS):
        raise TypeError("Expecting a 'CMOS' detector object.")

    signal_mean_array = detector.charge.array.astype("float64")
    signal_non_linear = compute_simple_physical_non_linearity(
        array_2d=signal_mean_array,
        temperature=detector.environment.temperature,
        v_bias=v_bias,
        cutoff=cutoff,
        n_acceptor=n_acceptor,
        n_donor=n_donor,
        diode_diameter=diode_diameter,
    )

    detector.signal.array = signal_non_linear


def compute_physical_non_linearity(
    array_2d: np.ndarray,
    temperature: float,
    fixed_capacitance: float,
    v_bias: float,
    cutoff: float,
    n_acceptor: float,
    n_donor: float,
    diode_diameter: float,
) -> np.ndarray:
    """Compute physical non-linear signal.

    Parameters
    ----------
    array_2d : ndarray
        Input array.
    temperature
        Temperature. Unit: K.
    fixed_capacitance : float
        Additional fixed capacitance. Unit: F
    v_bias : float
        Initial bias voltage. Unit: V.
    cutoff : float
        Cutoff wavelength. unit: um
    n_acceptor : float
        Acceptor density. Unit: atoms/cm^3
    n_donor : float
        Donor density. Unit: atoms/cm^3
    diode_diameter : float
        Diode diameter. Unit: um

    Returns
    -------
    ndarray
        Output array.
    """
    # Derivation of Cd concentration in the alloy,  it depends on cutoff wavelength and targeted operating temperature
    # Here we are considering the case where the detector is operated at its nominal temperature,
    # it might not be always the case
    e_g_targeted = 1.24 / cutoff  # cutoff is um and Eg in eV
    xcd = np.linspace(0.2, 0.6, 1000)
    targeted_operating_temperature = temperature
    e_g_calculated = hgcdte_bandgap(
        xcd, targeted_operating_temperature
    )  # Expected band-gap
    index = np.where(e_g_calculated > e_g_targeted)[0][0]
    x_cd = xcd[index]  # Targeted cadmium concentration in the HgCdTe alloy

    if not (0.2 <= x_cd <= 0.6):
        raise ValueError(
            "Hansen bangap expression used out of its nominal application range. "
            "x_cd must be between 0.2 and 0.6"
        )

    # Calculate the effective band-gap value at the temperature at which simulations are performed.
    ni = ni_hansen(x_cd=x_cd, temperature=temperature)

    # Build in potential
    vbi = (
        const.k_B.value
        * temperature
        / const.e.value
        * np.log(n_acceptor * n_donor / ni**2)
    )  # in V

    # HgCdTe dielectric constant
    eps = 20.5 - 15.6 * x_cd + 5.7 * x_cd**2  # without dimension

    # Surface of the diode, assumed to be planar
    surface = (
        np.pi * (diode_diameter / 2.0 * 1e-6) ** 2
    )  # Surface of the diode, assumed to be circular (in cm)

    # Initial value  of the diode capacitance
    co = surface * np.sqrt(
        (const.e.value * eps * const.eps0.value)
        / (2 * vbi)
        * ((n_acceptor * 1e6 * n_donor * 1e6) / (n_acceptor * 1e6 + n_donor * 1e6))
    )

    # Resolution of 2nd order equation
    b = -2.0 * co / fixed_capacitance
    a = -1.0
    c = (
        (1 - v_bias / vbi)
        + 2.0 * co / fixed_capacitance * np.sqrt(1.0 - v_bias / vbi)
        - array_2d * const.e.value / (fixed_capacitance * vbi)
    )
    discriminant = b**2 - 4.0 * c * a
    u1 = (-b - np.sqrt(discriminant)) / (2.0 * a)
    v1 = (1.0 - u1**2) * vbi - v_bias  # is subtracted it only deals with offset level

    array = np.copy(v1)  # unit if V, voltage at the gate of the pixel SFD
    non_linear_signal = array.astype("float64")

    return non_linear_signal


def physical_non_linearity(
    detector: Detector,
    cutoff: float,
    n_acceptor: float,
    n_donor: float,
    diode_diameter: float,
    v_bias: float,
    fixed_capacitance: float,
) -> None:
    """Apply physical non-linearity.

    Parameters
    ----------
    detector : Detector
        Pyxel detector object.
    cutoff : float
        Cutoff wavelength. unit: um
    n_donor : float
        Donor density. Unit: atoms/cm^3
    n_acceptor : float
        Acceptor density. Unit: atoms/cm^3
    diode_diameter : float
        Diode diameter. Unit: um
    v_bias : float
        Initial bias voltage. Unit: V.
    fixed_capacitance : float
        Additional fixed capacitance. Unit: F
    """
    if not (4 <= detector.environment.temperature <= 300):
        raise ValueError(
            "Hansen bangap expression used out of its nominal application range. "
            "temperature must be between 4K and 300K"
        )

    if not isinstance(detector, CMOS):
        raise TypeError("Expecting a 'CMOS' detector object.")

    signal_mean_array = detector.pixel.array.astype("float64")
    signal_non_linear = compute_physical_non_linearity(
        array_2d=signal_mean_array,
        temperature=detector.environment.temperature,
        fixed_capacitance=fixed_capacitance,
        v_bias=v_bias,
        cutoff=cutoff,
        n_acceptor=n_acceptor,
        n_donor=n_donor,
        diode_diameter=diode_diameter,
    )

    detector.signal.array = signal_non_linear


def compute_physical_non_linearity_with_saturation(
    signal_array_2d: np.ndarray,
    photon_array_2d: np.ndarray,
    time_step: float,
    step_number: int,
    temperature: float,
    cutoff: float,
    n_donor: float,
    n_acceptor: float,
    phi_implant: float,
    d_implant: float,
    saturation_current: float,
    ideality_factor: float,
    v_reset: float,
    d_sub: float,
    fixed_capacitance: float,
    euler_points: int,
) -> np.ndarray:
    """Compute physical non-linear signal with saturation.

    Parameters
    ----------
    signal_array_2d : ndarray
        Input signal array.
    photon_array_2d : ndarray
        Input photon array.
    time_step
        Time step. Unit: s.
    step_number : int
        Step number.
    temperature : float
        Temperature. Unit: K.
    cutoff : float
        Cutoff wavelength. unit: um.
    n_donor : float
        Donor density. Unit: atoms/cm^3.
    n_acceptor : float
        Acceptor density. Unit: atoms/cm^3.
    phi_implant : float
        Diameter of the implantation. Unit: um.
    d_implant : float
        Depth of the implamantation. Unit: um.
    saturation_current : float
        Saturation current: e-/s/pix..
    ideality_factor : float
        Ideality factor.
    v_reset : float
        VRESET. Unit: V.
    d_sub : float
        DSUB. Unit: V.
    fixed_capacitance : float
        Additional fixed capacitance. Unit: F.
    euler_points : int
        Number of points in the euler method.

    Returns
    -------
    ndarray
        Output array containing non-linear signal. Unit: V.
    """
    # Derivation of Cd concentration in the alloy,  it depends on cutoff wavelength and targeted operating temperature
    # Here we are considering the case where the detector is operated at its nominal temperature,
    # it might not be always the case
    e_g_targeted = 1.24 / cutoff  # cutoff is um and Eg in eV
    xcd = np.linspace(0.2, 0.6, 1000)
    # targeted_operating_temperature = temperature
    e_g_calculated = hgcdte_bandgap(xcd, temperature)  # Expected bandgap
    index = np.where(e_g_calculated > e_g_targeted)[0][0]
    x_cd = xcd[index]  # Targeted cadmium concentration in the HgCdTe alloy
    # Calculate the effective bandgap value at the temperature at which simulations are performed.

    if not (0.2 <= x_cd <= 0.6):
        raise ValueError(
            "Hansen bangap expression used out of its nominal application range. "
            "x_cd must be between 0.2 and 0.6"
        )

    row, col = signal_array_2d.shape

    phi_implant = phi_implant * 1e-6  # in m
    d_implant = d_implant * 1e-6  # in m

    if step_number == 0:
        signal_array_2d = v_reset * np.ones((row, col))

    # detector.signal.array should be expressed in unit of mV. It is the bias at the gate of the pixel SFD ????
    det_polar = euler(
        time_step=time_step,
        nb_pts=euler_points,
        v_bias=np.ravel(signal_array_2d) - d_sub,
        phi_implant=phi_implant,
        d_implant=d_implant,
        n_acceptor=n_acceptor,
        n_donor=n_donor,
        x_cd=x_cd,
        temperature=temperature,
        photonic_current=np.ravel(photon_array_2d) / time_step,
        fixed_capacitance=fixed_capacitance,
        sat_current=saturation_current,
        n=ideality_factor,
    )

    array = np.reshape(det_polar + d_sub, (row, col))
    non_linear_signal = array.astype("float64")

    return non_linear_signal


def physical_non_linearity_with_saturation(
    detector: Detector,
    cutoff: float,
    n_donor: float,
    n_acceptor: float,
    phi_implant: float,
    d_implant: float,
    saturation_current: float,
    ideality_factor: float,
    v_reset: float,
    d_sub: float,
    fixed_capacitance: float,
    euler_points: int,
) -> None:
    """Apply physical non-linearity with saturation.

    Parameters
    ----------
    detector : Detector
        Pyxel detector object.
    cutoff : float
        Cutoff wavelength. unit: um
    n_donor : float
        Donor density. Unit: atoms/cm^3
    n_acceptor : float
        Acceptor density. Unit: atoms/cm^3
    phi_implant : float
        Diameter of the implantation. Unit: um
    d_implant : float
        Depth of the implamantation. Unit: um
    saturation_current : float
        Saturation current: e-/s/pix.
    ideality_factor : float
        Ideality factor.
    v_reset : float
        VRESET. Unit: V.
    d_sub : float
        DSUB. Unit: V.
    fixed_capacitance : float
        Additional fixed capacitance. Unit: F.
    euler_points : int
        Number of points in the euler method.
    """
    if not (4 <= detector.environment.temperature <= 300):
        raise ValueError(
            "Hansen bangap expression used out of its nominal application range. "
            "temperature must be between 4K and 300K"
        )

    if not isinstance(detector, CMOS):
        raise TypeError("Expecting a 'CMOS' detector object.")

    if detector.is_first_readout:
        # This is the first step
        signal_2d = detector.signal.array
    else:
        signal_2d = np.array(detector.data["/non_linearity_with_saturation/previous"])

        if detector.is_last_readout:
            # This is the last step. Remove the previous value
            detector.data["/non_linearity_with_saturation"].orphan()

    signal_non_linear = compute_physical_non_linearity_with_saturation(
        signal_array_2d=signal_2d,
        photon_array_2d=detector.photon.array,
        time_step=detector.time_step,
        step_number=detector.pipeline_count,
        temperature=detector.environment.temperature,
        cutoff=cutoff,
        n_donor=n_donor,
        n_acceptor=n_acceptor,
        phi_implant=phi_implant,
        d_implant=d_implant,
        saturation_current=saturation_current,
        ideality_factor=ideality_factor,
        v_reset=v_reset,
        d_sub=d_sub,
        fixed_capacitance=fixed_capacitance,
        euler_points=euler_points,
    )

    # Update previous value
    detector.data["/non_linearity_with_saturation/previous"] = signal_non_linear

    detector.signal.array = signal_non_linear



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from pyxel.exposure import Readout


@pytest.mark.parametrize(
    "times, exp_times",
    [
        ([0.1], np.array([0.1])),
        ([1, 2, 3], np.array([1.0, 2.0, 3.0])),
        ("numpy.arange(0.5, 2, 0.5)", np.array([0.5, 1.0, 1.5])),
    ],
)
def test_readout_times(times, exp_times):
    """Test 'Readout' with 'times' parameters."""
    obj = Readout(times=times)

    assert_array_equal(obj.times, exp_times)


@pytest.mark.parametrize(
    "times",
    [[3, 2, 1], [1, 2, 3, 3, 4], [1, 3, 2, 4], "numpy.linspace(start=10, stop=1)"],
)
def test_readout_times_not_monotonic(times):
    """Test 'Readout' with value non-monotonic increasing."""
    with pytest.raises(ValueError, match=r"Readout times must be strictly increasing"):
        _ = Readout(times)


def test_readout_times_error_non_zero():
    """Test 'Readout' with times starting with 0."""
    with pytest.raises(ValueError, match=r"Readout times should be non-zero values"):
        _ = Readout(times=[0])


@pytest.mark.parametrize(
    "times, start_time",
    [
        ([1], 1),
        ([1], 2),
    ],
)
def test_readout_times_greater_than_start(times, start_time):
    """Test 'Readout' with times greater than start time."""
    with pytest.raises(
        ValueError, match=r"Readout times should be greater than start time"
    ):
        _ = Readout(times=times, start_time=start_time)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

""":term:`CDM` model calibration with PYGMO.

https://esa.github.io/pagmo2/index.html
"""
import copy
import logging
import math
from collections.abc import Sequence
from copy import deepcopy
from numbers import Number
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Optional, Union

import numpy as np
import pandas as pd
from dask.delayed import delayed

from pyxel.calibration import (
    CalibrationMode,
    FittingCallable,
    ProblemSingleObjective,
    check_ranges,
    list_to_3d_slice,
    list_to_slice,
    read_data,
    read_datacubes,
)
from pyxel.exposure import run_exposure_pipeline
from pyxel.observation import ParameterValues
from pyxel.pipelines import Processor, ResultId

if TYPE_CHECKING:
    import xarray as xr
    from numpy.typing import ArrayLike, NDArray

    from pyxel.exposure import Readout


class ModelFitting(ProblemSingleObjective):
    """Pygmo problem class to fit data with any model in Pyxel."""

    def __init__(
        self,
        processor: Processor,
        variables: Sequence[ParameterValues],
        readout: "Readout",
        calibration_mode: CalibrationMode,
        simulation_output: ResultId,
        generations: int,
        population_size: int,
        fitness_func: FittingCallable,
        file_path: Path,
        pipeline_seed: Optional[int] = None,
    ):
        self.processor: Processor = processor
        self.variables: Sequence[ParameterValues] = variables

        self.calibration_mode: CalibrationMode = calibration_mode
        self.original_processor: Optional[Processor] = None
        self.generations: int = generations
        self.pop: int = population_size
        self.readout: Readout = readout

        self.all_target_data: list[np.ndarray] = []
        self.weighting: Optional[np.ndarray] = None
        self.weighting_from_file: Optional[Sequence[np.ndarray]] = None
        self.fitness_func: FittingCallable = fitness_func
        self.sim_output: ResultId = simulation_output
        self.param_processor_list: list[Processor] = []

        self.file_path: Path = file_path
        self.pipeline_seed: Optional[int] = pipeline_seed

        self.fitness_array: Optional[np.ndarray] = None
        self.population: Optional[np.ndarray] = None
        self.champion_f_list: Optional[np.ndarray] = None
        self.champion_x_list: Optional[np.ndarray] = None

        self.lbd: list[float] = []  # lower boundary
        self.ubd: list[float] = []  # upper boundary

        self.sim_fit_range: tuple[slice, slice, slice] = (
            slice(None),
            slice(None),
            slice(None),
        )
        self.targ_fit_range: Union[tuple[slice, slice], tuple[slice, slice, slice]] = (
            slice(None),
            slice(None),
        )

        self.match: dict[int, list[str]] = {}

    def get_bounds(self) -> tuple[Sequence[float], Sequence[float]]:
        """Get the box bounds of the problem (lower_boundary, upper_boundary).

        It also implicitly defines the dimension of the problem.

        Returns
        -------
        tuple of lower boundaries and upper boundaries
        """
        return self.lbd, self.ubd

    def configure(
        self,
        target_fit_range: Sequence[int],
        out_fit_range: Sequence[int],
        target_output: Sequence[Path],
        input_arguments: Optional[Sequence[ParameterValues]] = None,
        weights: Optional[Sequence[float]] = None,
        weights_from_file: Optional[Sequence[Path]] = None,
    ) -> None:
        """TBW."""
        # if self.calibration_mode == 'single_model':           # TODO update
        #     self.single_model_calibration()

        self.set_bound()

        self.original_processor = deepcopy(self.processor)
        if input_arguments:
            max_val, min_val = 0, 1000
            for arg in input_arguments:
                min_val = min(min_val, len(arg.values))
                max_val = max(max_val, len(arg.values))
            if min_val != max_val:
                logging.warning(
                    'The "result_input_arguments" value lists have different lengths! '
                    "Some values will be ignored."
                )
            for i in range(min_val):
                new_processor: Processor = deepcopy(self.processor)

                step: ParameterValues
                for step in input_arguments:
                    assert step.values != "_"

                    value: Union[Literal["_"], str, Number] = step.values[i]

                    step.current = value
                    new_processor.set(key=step.key, value=step.current)
                self.param_processor_list += [new_processor]
        else:
            self.param_processor_list = [deepcopy(self.processor)]

        params: int = 0

        var: ParameterValues
        for var in self.variables:
            if isinstance(var.values, list):
                b = len(var.values)
            else:
                b = 1

            params += b
        self.champion_f_list = np.zeros((1, 1))
        self.champion_x_list = np.zeros((1, params))

        if self.readout.time_domain_simulation:
            target_list_3d: Sequence[np.ndarray] = read_datacubes(
                filenames=target_output
            )
            times, rows, cols = target_list_3d[0].shape
            check_ranges(
                target_fit_range=target_fit_range,
                out_fit_range=out_fit_range,
                rows=rows,
                cols=cols,
                readout_times=times,
            )
            self.targ_fit_range = list_to_slice(target_fit_range)
            self.sim_fit_range = list_to_3d_slice(out_fit_range)

            target_3d: np.ndarray
            for target_3d in target_list_3d:
                self.all_target_data += [target_3d[self.targ_fit_range]]

        else:
            target_list_2d: Sequence[np.ndarray] = read_data(filenames=target_output)

            rows, cols = target_list_2d[0].shape
            check_ranges(
                target_fit_range=target_fit_range,
                out_fit_range=out_fit_range,
                rows=rows,
                cols=cols,
            )
            self.targ_fit_range = list_to_slice(target_fit_range)

            # TODO: remove 'type: ignore' and fix the issue
            out_fit_range = [None, None, *out_fit_range]  # type: ignore
            self.sim_fit_range = list_to_3d_slice(out_fit_range)

            target_2d: np.ndarray
            for target_2d in target_list_2d:
                self.all_target_data += [target_2d[self.targ_fit_range]]

            self._configure_weights(
                weights=weights, weights_from_file=weights_from_file
            )

    def _configure_weights(
        self,
        weights: Optional[Sequence[float]] = None,
        weights_from_file: Optional[Sequence[Path]] = None,
    ) -> None:
        """TBW.

        Parameters
        ----------
        weights
        weights_from_file

        Returns
        -------
        None
        """
        if weights_from_file is not None:
            if self.readout.time_domain_simulation:
                wf = read_datacubes(weights_from_file)
                self.weighting_from_file = [
                    weight_array[self.targ_fit_range] for weight_array in wf
                ]
            else:
                wf = read_data(weights_from_file)
                self.weighting_from_file = [
                    weight_array[self.targ_fit_range] for weight_array in wf
                ]
        elif weights is not None:
            self.weighting = np.array(weights)

    def set_bound(self) -> None:
        """TBW."""
        self.lbd = []
        self.ubd = []

        var: ParameterValues
        for var in self.variables:
            assert var.boundaries is not None  # TODO: Fix this

            if var.values == "_":
                assert var.boundaries.shape == (2,)  # TODO: Fix this

                low_val: float
                high_val: float
                low_val, high_val = var.boundaries

                if var.logarithmic:
                    low_val = math.log10(low_val)
                    high_val = math.log10(high_val)

                self.lbd += [low_val]
                self.ubd += [high_val]

            elif isinstance(var.values, Sequence) and all(
                x == "_" for x in var.values[:]
            ):
                if var.boundaries.ndim == 1:
                    low_val, high_val = var.boundaries

                    low_values: NDArray[np.float_] = np.array(
                        [low_val] * len(var.values)
                    )
                    high_values: NDArray[np.float_] = np.array(
                        [high_val] * len(var.values)
                    )

                elif var.boundaries.ndim == 2:
                    low_values = var.boundaries[:, 0]
                    high_values = var.boundaries[:, 1]
                else:
                    raise NotImplementedError

                if var.logarithmic:
                    low_values = np.log10(low_values)
                    high_values = np.log10(high_values)

                self.lbd += low_values.tolist()
                self.ubd += high_values.tolist()

            else:
                raise ValueError(
                    'Character "_" (or a list of it) should be used to '
                    "indicate variables need to be calibrated"
                )

    def get_simulated_data(self, processor: Processor) -> np.ndarray:
        """Extract 2D data from a processor."""
        if self.sim_output not in ("image", "signal", "pixel"):
            raise NotImplementedError(
                f"Simulation mode: {self.sim_output!r} not implemented"
            )

        simulated_data: np.ndarray = processor.result[self.sim_output][
            self.sim_fit_range
        ]

        return simulated_data

    def calculate_fitness(
        self,
        simulated_data: np.ndarray,
        target_data: np.ndarray,
        weighting: Optional[np.ndarray] = None,
    ) -> float:
        """TBW.

        Parameters
        ----------
        simulated_data
        target_data
        weighting
        """
        if weighting is not None:
            factor = weighting
        else:
            factor = np.ones(np.shape(target_data))

        fitness: float = self.fitness_func(
            simulated=simulated_data.astype(np.float64),
            target=target_data.astype(np.float64),
            weighting=factor.astype(np.float64),
        )

        return fitness

    # TODO: If possible, use 'numba' for this method
    def fitness(self, decision_vector_1d: np.ndarray) -> Sequence[float]:
        """Call the fitness function, elements of parameter array could be logarithmic values.

        Parameters
        ----------
        decision_vector_1d : array_like
            A 1d decision vector.

        Returns
        -------
        sequence
            The fitness of the input decision vector (concatenating the objectives,
            the equality and the inequality constraints)
        """
        # TODO: Fix this
        if self.pop is None:
            raise NotImplementedError("'pop' is not initialized.")

        try:
            # TODO: Use directory 'logging.'
            logger = logging.getLogger("pyxel")
            prev_log_level = logger.getEffectiveLevel()

            parameter_1d = self.convert_to_parameters(decision_vector_1d)
            # TODO: deepcopy is not needed. Check this
            processor_list: Sequence[Processor] = self.param_processor_list

            overall_fitness: float = 0.0
            for i, (processor, target_data) in enumerate(
                zip(processor_list, self.all_target_data)
            ):
                processor = self.update_processor(
                    parameter=parameter_1d, processor=processor
                )

                logger.setLevel(logging.WARNING)
                # result_proc = None
                if self.calibration_mode == CalibrationMode.Pipeline:
                    _ = run_exposure_pipeline(
                        processor=processor,
                        readout=self.readout,
                        pipeline_seed=self.pipeline_seed,
                    )
                # elif self.calibration_mode == 'single_model':
                #     self.fitted_model.function(processor.detector)               # todo: update
                else:
                    raise NotImplementedError

                logger.setLevel(prev_log_level)

                simulated_data = self.get_simulated_data(processor=processor)

                weighting: Optional[np.ndarray] = None

                if self.weighting is not None:
                    weighting = self.weighting[i] * np.ones(
                        (
                            processor.detector.geometry.row,
                            processor.detector.geometry.col,
                        )
                    )
                elif self.weighting_from_file is not None:
                    weighting = self.weighting_from_file[i]

                # TODO: Create a multi-objective problem and use
                #       pygmo.decompose to convert it into a single-objective problem
                overall_fitness += self.calculate_fitness(
                    simulated_data=simulated_data,
                    target_data=target_data,
                    weighting=weighting,
                )

        except Exception:
            logging.exception(
                "Catch an exception in 'fitness' for ModelFitting: %r.", self
            )
            raise

        return [overall_fitness]

    def convert_to_parameters(self, decisions_vector: "ArrayLike") -> np.ndarray:
        """Convert a decision version from Pygmo2 to parameters.

        Parameters
        ----------
        decisions_vector : array_like
            It could a 1D or 2D array.

        Returns
        -------
        array_like
            Parameters
        """
        parameters = np.array(decisions_vector)

        a = 0
        for var in self.variables:
            b = 1
            if isinstance(var.values, list):
                b = len(var.values)
            if var.logarithmic:
                start = a
                stop = a + b
                parameters[..., start:stop] = np.power(10, parameters[..., start:stop])
            a += b

        return parameters

    def apply_parameters(
        self, processor: Processor, parameter: np.ndarray
    ) -> Processor:
        """Create a new ``Processor`` with new parameters."""
        new_processor = self.update_processor(parameter=parameter, processor=processor)

        _ = run_exposure_pipeline(
            processor=new_processor,
            readout=self.readout,
            pipeline_seed=self.pipeline_seed,
        )

        return new_processor

    def apply_parameters_to_processors(
        self, parameters: "xr.DataArray"
    ) -> pd.DataFrame:
        """TBW."""
        assert "island" in parameters.dims
        assert "param_id" in parameters.dims

        lst = []
        for id_processor, processor in enumerate(self.param_processor_list):
            delayed_processor = delayed(processor)

            for idx_island, params_array in parameters.groupby("island"):
                params: np.ndarray = params_array.data  # type

                result_processor = delayed(self.apply_parameters)(
                    processor=delayed_processor, parameter=params
                )

                lst.append(
                    {
                        "island": idx_island,
                        "id_processor": id_processor,
                        "processor": result_processor,
                    }
                )

        df = pd.DataFrame(lst).sort_values(["island", "id_processor"])

        return df

    def update_processor(
        self, parameter: np.ndarray, processor: Processor
    ) -> Processor:
        """TBW."""
        new_processor = copy.deepcopy(processor)
        a, b = 0, 0
        for var in self.variables:
            if var.values == "_":
                b = 1
                new_processor.set(key=var.key, value=parameter[a])
            elif isinstance(var.values, list):
                b = len(var.values)

                start = a
                stop = a + b
                new_processor.set(key=var.key, value=parameter[start:stop])
            a += b
        return new_processor



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


"""Tests for dark current models."""

import pytest

from pyxel.detectors import (
    CCD,
    CMOS,
    CCDGeometry,
    Characteristics,
    CMOSGeometry,
    Environment,
    ReadoutProperties,
)
from pyxel.models.charge_generation import dark_current_rule07


@pytest.fixture
def cmos_10x10() -> CMOS:
    """Create a valid CCD detector."""
    detector = CMOS(
        geometry=CMOSGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(temperature=273.15),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


def test_dark_current_rule07_valid(cmos_10x10: CMOS):
    """Test model 'dark_current_rule07' with valid inputs."""
    dark_current_rule07(
        detector=cmos_10x10,
    )


def test_dark_current_rule07_with_ccd():
    """Test model 'dark_current_rule07' with a `CCD` detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )

    with pytest.raises(TypeError, match="Expecting a CMOS object for detector."):
        dark_current_rule07(
            detector=detector,
        )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import pytest

from pyxel.detectors import CCD, CCDGeometry, Characteristics, Environment
from pyxel.models.charge_transfer import arctic_remove


@pytest.fixture
def ccd_10x10() -> CCD:
    """Create a valid CCD detector."""
    return CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )


@pytest.mark.parametrize(
    "num_iterations, instant_traps, exp_exc, exp_msg",
    [
        pytest.param(
            0,
            [{"density": 1.0, "release_timescale": 10.0}],
            ValueError,
            "Number of iterations must be > 1",
            id="num_iterations == 0",
        ),
        pytest.param(
            -1,
            [{"density": 1.0, "release_timescale": 10.0}],
            ValueError,
            "Number of iterations must be > 1",
            id="num_iterations < 0",
        ),
        pytest.param(
            1,
            [{"release_timescale": 10.0}],
            KeyError,
            "Missing key 'density' in parameter 'instant_traps'",
            id="num_iterations < 0",
        ),
        pytest.param(
            1,
            [{"density": 1.0}],
            KeyError,
            "Missing key 'release_timescale' in parameter 'instant_traps'",
            id="num_iterations < 0",
        ),
    ],
)
def test_arctic_remove_bad_inputs(
    ccd_10x10: CCD, num_iterations, instant_traps, exp_exc, exp_msg
):
    """Test function 'arctic_remove' with bad inputs."""
    with pytest.raises(exp_exc, match=exp_msg):
        arctic_remove(
            detector=ccd_10x10,
            well_fill_power=10.0,
            instant_traps=instant_traps,
            num_iterations=num_iterations,
            express=0,
        )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


import pytest

from pyxel.detectors import CCD, CCDGeometry, Characteristics, Environment
from pyxel.models.readout_electronics import simple_phase_conversion


def test_simple_phase_conversion_with_ccd():
    """Test model 'dead_time_filter' with a `CCD` detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=5,
            col=4,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )

    with pytest.raises(TypeError, match="Expecting a MKID object for the detector."):
        simple_phase_conversion(detector=detector)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import pytest

from pyxel.detectors import (
    CCD,
    CMOS,
    CCDGeometry,
    Characteristics,
    CMOSGeometry,
    Environment,
    ReadoutProperties,
)
from pyxel.models.charge_collection import simple_persistence


@pytest.fixture
def cmos_5x10() -> CMOS:
    """Create a valid CCD detector."""
    detector = CMOS(
        geometry=CMOSGeometry(
            row=5,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])

    return detector


@pytest.fixture
def ccd_5x5() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    return detector


def test_simple_persistence_two_traps(cmos_5x10: CMOS):
    """Test model 'simple_persistence'."""
    detector = cmos_5x10

    # No persistence
    assert not detector.has_persistence()

    simple_persistence(
        detector=detector,
        trap_time_constants=[1.0, 3.0],
        trap_densities=[0.1, 0.1],
    )

    assert detector.has_persistence()
    assert len(detector.persistence.trap_list) == 2


def test_simple_persistence_one_trap(cmos_5x10: CMOS):
    """Test model 'simple_persistence'."""
    detector = cmos_5x10

    # No persistence
    assert not detector.has_persistence()

    # With persistence
    simple_persistence(
        detector=detector,
        trap_time_constants=[1.0],
        trap_densities=[0.1],
    )

    assert detector.has_persistence()
    assert len(detector.persistence.trap_list) == 1


@pytest.mark.parametrize(
    "trap_time_constants, trap_densities, exp_error, exp_msg",
    [
        pytest.param(
            [],
            [],
            ValueError,
            "Expecting at least one 'trap_time_constants' and 'trap_densities'",
            id="no elements",
        ),
        pytest.param(
            [1.0],
            [0.1, 0.1],
            ValueError,
            "Expecting same number of elements for parameters",
            id="not same number of elements",
        ),
    ],
)
def test_simple_persistence_bad_inputs(
    cmos_5x10: CMOS,
    trap_time_constants,
    trap_densities,
    exp_error,
    exp_msg,
):
    """Test model 'simple_persistence' with bad inputs."""
    detector = cmos_5x10

    with pytest.raises(exp_error, match=exp_msg):
        simple_persistence(
            detector=detector,
            trap_time_constants=trap_time_constants,
            trap_densities=trap_densities,
        )


def test_persistence_with_ccd(ccd_5x5: CCD):
    """Test model 'simple_persistence' with a `CCD` detector."""
    detector = ccd_5x5

    with pytest.raises(TypeError, match="Expecting a CMOS object for detector."):
        simple_persistence(
            detector=detector,
            trap_time_constants=[1.0, 10.0],
            trap_densities=[0.1, 0.1],
        )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from collections.abc import Sequence
from typing import Literal, Optional

import pytest

from pyxel.detectors import (
    CCD,
    CCDGeometry,
    Characteristics,
    Environment,
    ReadoutProperties,
)
from pyxel.models.photon_collection import illumination


@pytest.fixture
def ccd_10x10() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


@pytest.mark.parametrize(
    "level, option, object_size, object_center, time_scale",
    [
        pytest.param(1.0, "uniform", None, None, 1.0, id="valid_uniform"),
        pytest.param(1.0, "elliptic", [2, 2], [5, 5], 1.0, id="valid_elliptic"),
        pytest.param(1.0, "rectangular", [2, 2], [5, 5], 1.0, id="valid_rectangular"),
        pytest.param(
            1.0,
            "rectangular",
            [2, 2],
            [15, 15],
            1.0,
            id="invalid_center_rectangular",
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
        ),
        pytest.param(
            1.0,
            "elliptic",
            [2, 2],
            [15, 15],
            1.0,
            id="invalid_center_elliptic",
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
        ),
        pytest.param(
            1.0,
            "elliptic",
            [2, 2],
            "foo",
            1.0,
            id="invalid_center_elliptic",
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
        ),
        pytest.param(
            1.0,
            "rectangular",
            [2, 2],
            "foo",
            1.0,
            id="invalid_center_rectangular",
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
        ),
    ],
)
def test_illumination(
    ccd_10x10: CCD,
    level: float,
    option: Literal["uniform", "rectangular", "elliptic"],
    object_size: Optional[Sequence[int]],
    object_center: Optional[Sequence[int]],
    time_scale: float,
):
    """Test input parameters for function 'illumination'."""
    illumination(
        detector=ccd_10x10,
        level=level,
        option=option,
        object_size=object_size,
        object_center=object_center,
        time_scale=time_scale,
    )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


from pathlib import Path

import pytest

from pyxel.models.charge_generation.charge_deposition import simulate_charge_deposition


def test_simulate_charge_deposition_number_of_particles() -> None:
    flux = 0
    exposure = 1
    x_lim = 50
    y_lim = 50
    z_lim = 50
    step_size = 1
    energy_mean = 1
    energy_spread = 1
    energy_spectrum = None
    energy_spectrum_sampling = "log"
    ehpair_creation = 3.65
    material_density = 2.3290
    particle_direction = None
    stopping_power_curve = "data/mct-stopping-power.csv"

    # Run model
    with pytest.raises(
        ValueError, match="The number of particles generated has to be greater than 0."
    ):
        simulate_charge_deposition(
            flux=flux,
            exposure=exposure,
            x_lim=x_lim,
            y_lim=y_lim,
            z_lim=z_lim,
            step_size=step_size,
            energy_mean=energy_mean,
            energy_spread=energy_spread,
            energy_spectrum=energy_spectrum,
            energy_spectrum_sampling=energy_spectrum_sampling,
            ehpair_creation=ehpair_creation,
            material_density=material_density,
            particle_direction=particle_direction,
            stopping_power_curve=stopping_power_curve,
        )


@pytest.mark.parametrize("x_lim,y_lim,z_lim", [(-1, 2, 4), (2, -1, 4), (2, 1, -4)])
def test_simulate_charge_deposition_sampling_argument(x_lim, y_lim, z_lim) -> None:
    flux = 1
    exposure = 1
    step_size = 1
    energy_mean = 1
    energy_spread = 1
    energy_spectrum = None
    energy_spectrum_sampling = 1
    ehpair_creation = 3.65
    material_density = 2.3290
    particle_direction = None
    stopping_power_curve = Path(
        "pyxel/models/charge_generation/data/mct-stopping-power.csv"
    ).resolve()

    # Run model
    with pytest.raises(ValueError, match="Detector dimension is negative or 0."):
        simulate_charge_deposition(
            flux=flux,
            exposure=exposure,
            x_lim=x_lim,
            y_lim=y_lim,
            z_lim=z_lim,
            step_size=step_size,
            energy_mean=energy_mean,
            energy_spread=energy_spread,
            energy_spectrum=energy_spectrum,
            energy_spectrum_sampling=energy_spectrum_sampling,
            ehpair_creation=ehpair_creation,
            material_density=material_density,
            particle_direction=particle_direction,
            stopping_power_curve=str(stopping_power_curve),
        )



# Copyright (c) 2023, Florian MORIOUSEF, Vincent GOIFFON, Alexandre LE ROCH, Aubin ANTONSANTI, ISAE-SUPAERO
#
# vincent.goiffon@isae-supaero.fr
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


"""Model to generate charge due to dark current induced by radiation.

The Dark Current Model description can be found in:
A. Le Roch et al., "Radiation-Induced Leakage Current and Electric Field Enhancement in CMOS Image Sensor Sense Node
Floating Diffusions,"
in IEEE Transactions on Nuclear Science, vol. 66, no. 3, pp. 616-624, March 2019, doi: 10.1109/TNS.2019.2892645.
Jean-Marc Belloir, Vincent Goiffon, Cédric Virmontois, Mélanie Raine, Philippe Paillet, Olivier Duhamel,
Marc Gaillardin, Romain Molina, Pierre Magnan, and Olivier Gilard, "Pixel pitch and particle energy influence on the
dark current distribution of neutron irradiated CMOS image sensors," Opt. Express 24, 4299-4315 (2016)
"""

import warnings
from typing import Optional

import numpy as np
from astropy import constants as const

from pyxel.detectors import CMOS
from pyxel.util import set_random_seed


def damage_factors(
    annealing_time: float,
    eact_dc: float,
    temperature: float,
    kdark_srour: float,
    gamma_dark: float,
    depletion_volume: float,
    displacement_dose: float,
) -> tuple[float, float]:
    """Calculaye damage factors and return as list [nu_dark, mu_dark].

    Parameters
    ----------
    annealing_time : float
        Annealing time. Unit: s
    eact_dc : float
        Activation energy parameter. Unit: eV
    temperature : float
        Temperature parameter. Unit K
    kdark_srour : float
        Damage factor K. Unit e-/cm3/sec per MeV/g
    gamma_dark : float
        Gamma dark parameter. Unit 1/µm3/(TeV/g)
    depletion_volume : float
        Depletion volume parameter. Unit µm3
    displacement_dose : float
        Displacement dose parameter. Unit TeV/g

    Returns
    -------
    float
        nu_dark. Unit: e-/s
    float
        mu_dark: Mean number of interactions per pixel. Unit: no units
    """

    k = const.k_B.value
    q = const.e.value

    # Linear fit of Srour & Lo annealing factor (TNS 2000) in the 10^4 - 5.10^6 s range
    annealing_factor = (-0.3965 * np.log10(annealing_time) + 3.5948) / 1.3024
    operating_temperature_correction = np.exp(
        (-eact_dc * q / k / temperature) - (-eact_dc * q / k / 300)
    )
    kdark = (
        kdark_srour
        * annealing_factor
        * operating_temperature_correction
        / 1e4**3
        * 1e6
    )
    nu_dark = kdark / gamma_dark
    mu_dark = (
        gamma_dark * depletion_volume * displacement_dose
    )  # mean number of interactions per pixel

    return nu_dark, mu_dark


def damage_factors_silicon(
    annealing_time: float,
    temperature: float,
    depletion_volume: float,
    displacement_dose: float,
) -> tuple[float, float]:
    """Return damage factors (for silicon device) calculation list [nu_dark, mu_dark].

    based on:
    Source: J. R. Srour and D. H. Lo, "Universal damage factor for radiation-induced dark current in silicon devices,"
    in IEEE Transactions     xon Nuclear Science, vol. 47, no. 6, pp. 2451-2459, Dec. 2000, doi: 10.1109/23.903792.

    Parameters
    ----------
    annealing_time : float
        Annealing time. Unit: weeks
    temperature :
        Temperature parameter. Unit: K
    depletion_volume :
        Depletion volume parameter. Unit: µm3
    displacement_dose :
        Displacement dose parameter. Unit: TeV/g

    Returns
    -------
    float
        nu_dark. Unit: e-/s
    float
        mu_dark: Mean number of interactions per pixel. Unit: no units
    """

    eact_dc = 0.63  # eV
    kdark_srour = 1.9e5  # e-/cm3/sec per MeV/g
    gamma_dark = 0.097179425484859 / 4100  # 1/µm3/(TeV/g)
    annealing_time = annealing_time * 7 * 24 * 3600  # convert week --> sec

    return damage_factors(
        annealing_time=annealing_time,
        eact_dc=eact_dc,
        temperature=temperature,
        kdark_srour=kdark_srour,
        gamma_dark=gamma_dark,
        depletion_volume=depletion_volume,
        displacement_dose=displacement_dose,
    )


def compute_radiation_induced_dark_current(
    number_of_rows: int,
    number_of_columns: int,
    mu_dark: float,
    nu_dark: float,
    integration_time: float,
    shot_noise: bool,
) -> np.ndarray:
    """Return Dark Signal Frame.

    Parameters
    ----------
    number_of_rows : int
        Number of rows. Unit: pixel
    number_of_columns : int
        Number of columns. Unit: pixel
    mu_dark : float
        Mean number of interactions per pixel. Unit: no unit
    nu_dark : float
        Parameter nu_dark. Unit: e-/s
    integration_time : float
        Integration time. Unit: s
    shot_noise : bool
        Shot noise: true for shotnoise

    Returns
    -------
    np.ndarray
        DarkCurrentFrame. Unit: e-
    """

    # Define the Dark Current Frame
    dark_current_frame = np.zeros((number_of_rows, number_of_columns))  # e-

    # Assign a number of interactions for each pixel of the frame
    interactions_array = np.random.poisson(
        mu_dark, size=(number_of_rows, number_of_columns)
    )

    for i in range(1, np.max(interactions_array)):
        # Index of pixels with i events
        temp = np.array(np.where(interactions_array == i))
        temp = np.swapaxes(temp, 0, 1)

        temp_darkcurrent = np.random.exponential(
            scale=nu_dark, size=(i, len(temp))
        )  # e-/s

        if i > 1:
            # if one pixel has more than 1 interaction, sum darkcurrent contributions
            temp_darkcurrent = np.sum(temp_darkcurrent, axis=0)
        else:
            temp_darkcurrent = np.reshape(
                temp_darkcurrent, (np.shape(temp_darkcurrent)[1])
            )

        for pixels in range(len(temp)):
            row = temp[pixels][0]
            column = temp[pixels][1]

            # assign dark current value to corresponding pixel, e-/s
            dark_current_frame[row, column] = temp_darkcurrent[pixels]

    dark_signal_frame = np.round(dark_current_frame * integration_time)  # e-

    if shot_noise:
        dark_signal_frame = np.random.poisson(dark_signal_frame).astype(float)

    if np.isinf(dark_signal_frame).any():
        warnings.warn(
            "Unphysical high value for dark current from fixed pattern noise distribution"
            " will result in inf values. Enable a FWC model to ensure a physical limit.",
            RuntimeWarning,
            stacklevel=2,
        )

    return dark_signal_frame


def radiation_induced_dark_current(
    detector: CMOS,
    depletion_volume: float,
    annealing_time: float,
    displacement_dose: float,
    shot_noise: bool,
    seed: Optional[int] = None,
) -> None:
    """Model to add dark current induced by radiation to the detector charge.

    The radiation induced dark current model description can be found in :cite:p:`RadiationLeRoch2019`
    and :cite:p:`Belloir:16`.

    Parameters
    ----------
    detector : Detector
        Pyxel detector object.
    depletion_volume : float
        Depletion volume parameter. Unit: µm3.
    annealing_time : float
        Annealing time. Unit: weeks
    displacement_dose : float
        Displacement dose parameter. Unit: TeV/g
    shot_noise : bool
        True to enable shotnoise.
    seed : int, optional
    """
    geo = detector.geometry
    temperature = detector.environment.temperature
    nu_dark, mu_dark = damage_factors_silicon(
        annealing_time=annealing_time,
        temperature=temperature,
        depletion_volume=depletion_volume,
        displacement_dose=displacement_dose,
    )

    integration_time = detector.time_step
    number_of_rows, number_of_columns = geo.shape

    with set_random_seed(seed):
        dark_signal_frame = compute_radiation_induced_dark_current(
            number_of_rows=number_of_rows,
            number_of_columns=number_of_columns,
            mu_dark=mu_dark,
            nu_dark=nu_dark,
            integration_time=integration_time,
            shot_noise=shot_noise,
        )
    detector.charge.add_charge_array(dark_signal_frame)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
from typing import Optional

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from pyxel.detectors.readout_properties import ReadoutProperties


@pytest.mark.parametrize(
    "times, start_time, exp_times, exp_steps, exp_is_linear",
    [
        pytest.param(
            [0.1], None, np.array([0.1]), np.array([0.1]), True, id="One element"
        ),
        pytest.param(
            [1, 2, 3],
            None,
            np.array([1, 2, 3], dtype=float),
            np.array([1, 1, 1], dtype=float),
            True,
            id="Three elements",
        ),
    ],
)
def test_readout_properties(
    times, start_time: Optional[float], exp_times, exp_steps, exp_is_linear
):
    """Test class 'ReadoutProperties'."""
    if start_time is None:
        obj = ReadoutProperties(times=times)
    else:
        obj = ReadoutProperties(times=times, start_time=start_time)

    result_times = obj.times
    result_steps = obj.steps
    result_times_linear = obj.times_linear

    assert_array_equal(result_times, exp_times)
    assert_array_equal(result_steps, exp_steps)
    assert_array_equal(result_times_linear, exp_is_linear)


@pytest.mark.parametrize(
    "times, start_time, exp_error, exp_msg",
    [
        pytest.param(
            [[1, 2, 3]], 0.0, ValueError, r"Readout times must be 1D", id="Not 1D"
        ),
        pytest.param(
            [3, 2, 1],
            0.0,
            ValueError,
            r"Readout times must be strictly increasing",
            id="Not monotonic",
        ),
        pytest.param(
            [0],
            0.0,
            ValueError,
            r"Readout times should be non-zero values",
            id="Non-zero values",
        ),
        pytest.param(
            [1],
            1,
            ValueError,
            r"Readout times should be greater than start time",
            id="Time greater than start time 1",
        ),
        pytest.param(
            [1],
            2,
            ValueError,
            r"Readout times should be greater than start time",
            id="Time greater than start time 1",
        ),
    ],
)
def test_readout_properties_wrong(times, start_time, exp_error, exp_msg):
    """Test class 'ReadoutProperties' with wrong inputs."""
    with pytest.raises(exp_error, match=exp_msg):
        _ = ReadoutProperties(times=times, start_time=start_time)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from collections import abc
from pathlib import Path

import pytest

import pyxel
from pyxel import Configuration
from pyxel.detectors import CCD
from pyxel.exposure import run_exposure_pipeline
from pyxel.observation import Observation, ParameterMode
from pyxel.pipelines import DetectionPipeline, Processor

expected_sequential = [
    (0, [("level", 10), ("initial_energy", 100)]),
    (1, [("level", 20), ("initial_energy", 100)]),
    (2, [("level", 30), ("initial_energy", 100)]),
    (3, [("level", 100), ("initial_energy", 100)]),
    (4, [("level", 100), ("initial_energy", 200)]),
    (5, [("level", 100), ("initial_energy", 300)]),
]

expected_product = [
    (0, [("level", 10), ("initial_energy", 100)]),
    (1, [("level", 10), ("initial_energy", 200)]),
    (2, [("level", 10), ("initial_energy", 300)]),
    (3, [("level", 20), ("initial_energy", 100)]),
    (4, [("level", 20), ("initial_energy", 200)]),
    (5, [("level", 20), ("initial_energy", 300)]),
    (6, [("level", 30), ("initial_energy", 100)]),
    (7, [("level", 30), ("initial_energy", 200)]),
    (8, [("level", 30), ("initial_energy", 300)]),
]


@pytest.mark.parametrize(
    "mode, expected",
    [
        # ('single', expected_single),
        (ParameterMode.Sequential, expected_sequential),
        (ParameterMode.Product, expected_product),
    ],
)
def test_pipeline_parametric_without_init_photon(mode: ParameterMode, expected):
    input_filename = "tests/data/parametric.yaml"
    cfg = pyxel.load(Path(input_filename))

    assert isinstance(cfg, Configuration)
    assert hasattr(cfg, "observation")
    assert hasattr(cfg, "ccd_detector")
    assert hasattr(cfg, "pipeline")

    observation = cfg.observation
    assert isinstance(observation, Observation)

    observation.parameter_mode = mode

    detector = cfg.ccd_detector
    assert isinstance(detector, CCD)

    pipeline = cfg.pipeline
    assert isinstance(pipeline, DetectionPipeline)

    processor = Processor(detector=detector, pipeline=pipeline)
    result = observation.debug_parameters(processor)
    assert result == expected

    processor_generator = observation._processors_it(processor=processor)
    assert isinstance(processor_generator, abc.Generator)

    for proc, _, _ in processor_generator:
        assert isinstance(proc, Processor)

        run_exposure_pipeline(processor=proc, readout=observation.readout)


@pytest.mark.deprecated
@pytest.mark.parametrize(
    "mode, expected",
    [
        # ('single', expected_single),
        (ParameterMode.Sequential, expected_sequential),
        (ParameterMode.Product, expected_product),
    ],
)
def test_pipeline_parametric_without_init_photon_deprecated(
    mode: ParameterMode, expected
):
    input_filename = "tests/data/deprecated_parametric.yaml"
    cfg = pyxel.load(Path(input_filename))

    assert isinstance(cfg, Configuration)
    assert hasattr(cfg, "observation")
    assert hasattr(cfg, "ccd_detector")
    assert hasattr(cfg, "pipeline")

    observation = cfg.observation
    assert isinstance(observation, Observation)

    observation.parameter_mode = mode

    detector = cfg.ccd_detector
    assert isinstance(detector, CCD)

    pipeline = cfg.pipeline
    assert isinstance(pipeline, DetectionPipeline)

    processor = Processor(detector=detector, pipeline=pipeline)
    result = observation.debug_parameters(processor)
    assert result == expected

    processor_generator = observation._processors_it(processor=processor)
    assert isinstance(processor_generator, abc.Generator)

    for proc, _, _ in processor_generator:
        assert isinstance(proc, Processor)

        run_exposure_pipeline(processor=proc, readout=observation.readout)



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#  #
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.

"""Tests for radiation induced dark current model."""

import pytest

from pyxel.detectors import (
    CMOS,
    Characteristics,
    CMOSGeometry,
    Environment,
    ReadoutProperties,
)
from pyxel.models.charge_generation import radiation_induced_dark_current


@pytest.fixture
def cmos_5x5() -> CMOS:
    """Create a valid CMOS detector."""
    detector = CMOS(
        geometry=CMOSGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(temperature=80),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


def test_radiation_induced_dark_current_valid(cmos_5x5: CMOS):
    """Test model 'radiation_induced_dark_current' with valid inputs."""
    detector = cmos_5x5

    # Call the function with appropriate arguments
    depletion_volume = 64  # µm3
    annealing_time = 0.1  # weeks
    displacement_dose = 50  # TeV/g
    seed = 42

    radiation_induced_dark_current(
        detector=detector,
        depletion_volume=depletion_volume,
        annealing_time=annealing_time,
        displacement_dose=displacement_dose,
        seed=seed,
        shot_noise=False,
    )



# Copyright or © or Copr. Thibault Pichon, CEA Paris-Saclay (2023)
#
# thibault.pichon@cea.fr
#
# This file is part of the Pyxel general simulator framework.
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
Non-linearity models utility functions.

@author: tpichon
"""
import numpy as np
from numba import float64, njit
from numba.experimental import jitclass

spec = [
    ("m_electron", float64),
    ("k_b", float64),
    ("electron_charge", float64),
    ("eps_0", float64),
]


@jitclass(spec)
class Constants:
    """Constants class."""

    def __init__(self):
        self.m_electron = 9.10938356e-31  # kg
        self.k_b = 1.38064852e-23  #
        self.electron_charge = 1.60217662e-19
        self.eps_0 = 8.85418782e-12


@njit
def build_in_potential(
    temperature: float,
    n_acceptor: float,
    n_donor: float,
    n_intrinsic: float,
) -> float:
    """Built-in potential.

    Parameters
    ----------
    temperature : float
        Temperature.
    n_acceptor
        Acceptor concentration.
    n_donor
        Donor concentration.
    n_intrinsic : float
        Intrinsic carrier concentration.

    Returns
    -------
    float
    """

    const = Constants()

    return (
        const.k_b
        * temperature
        / const.electron_charge
        * np.log(n_acceptor * n_donor / n_intrinsic**2)
    )


@njit
def w_dep(
    v_bias: np.ndarray,
    epsilon: float,
    n_acceptor: float,
    n_donor: float,
    x_cd: float,
    temperature: float,
) -> np.ndarray:
    """Depletion width.

    Parameters
    ----------
    v_bias : ndarray
        Bias voltage.
    epsilon : float
        Dielectric constant of the material.
    n_acceptor : float
        Acceptor concentration.
    n_donor : float
        Donor concentration.
    x_cd
        Cadmium composition between 0 and 1.
    temperature : float
        Temperature.
    """

    const = Constants()

    # Calculation of build in potential
    v_bi = build_in_potential(
        temperature=temperature,
        n_acceptor=n_acceptor,
        n_donor=n_donor,
        n_intrinsic=ni_hansen(x_cd=x_cd, temperature=temperature),
    )

    return np.sqrt(
        2.0
        * epsilon
        * const.eps_0
        / const.electron_charge
        * (n_acceptor * 1e6 + n_donor * 1e6)
        / (n_acceptor * 1e6 * n_donor * 1e6)
        * (v_bi - v_bias)
    )


@njit
def hgcdte_bandgap(x_cd: float, temperature: float) -> float:
    """Band gap energy in HgCdTe.

    This expression of the Gap of HgCdTe is valid for a Cadmium concentration between 0.2 and 0.6.
    Over a wide range of temperature between 4K and 300K

    Ref : Hansen, G. L., Schmit, J. L., & Casselman, T. N. (1982).
          Energy gap versus alloy composition and temperature in Hg1− x Cd x Te.
          Journal of Applied Physics, 53(10), 7099-7101.

    Parameters
    ----------
    x_cd : float
        Cadmium composition between 0 and 1
    temperature : float
        Temperature.

    Returns
    -------
    float
        Bandgap energy. Unit: eV
    """

    return (
        -0.302
        + 1.93 * x_cd
        - 0.81 * x_cd**2
        + 0.832 * x_cd**3
        + 5.35 * 1e-4 * (1 - 2 * x_cd) * temperature
    )


@njit
def ni_hansen(x_cd: float, temperature: float) -> float:
    """Intrinsic carrier concentration for HgCdTe.

    Ref : G.L. Hansen and J.L. Schmit  J. Applied Physics, 54, 1639 (1983)

    Parameters
    ----------
    x_cd : float
        Cadmium composition between 0 and 1
    temperature : float
        Temperature.

    Returns
    -------
    float
        intrinsic carrier concentration
    """

    const = Constants()

    e_g = hgcdte_bandgap(x_cd=x_cd, temperature=temperature)

    return (
        (
            5.585
            - 3.820 * x_cd
            + 1.753 * 1e-3 * temperature
            - 1.364 * 1e-3 * temperature * x_cd
        )
        * 1e14
        * e_g**0.75
        * temperature**1.5
        * np.exp(-const.electron_charge * e_g / (2 * const.k_b * temperature))
    )


@njit
def capa_pn_junction_cylindrical(
    v_bias: np.ndarray,
    phi_implant: float,
    d_implant: float,
    n_acceptor: float,
    n_donor: float,
    x_cd: float,
    temperature: float,
) -> np.ndarray:
    """Calculate the capacitance of the cylindrical pn junction.

    Parameters
    ----------
    v_bias : ndarray
    phi_implant : float
    d_implant : float
    n_acceptor : float
    n_donor : float
    x_cd : float
    temperature : float

    Returns
    -------
    ndarray
    """
    epsilon = (
        20.5 - 15.6 * x_cd + 5.7 * x_cd**2
    )  # Static dielectric constant, this value is ok for HgCdTe

    # Definition of the constants
    const = Constants()
    eps0 = const.eps_0

    # Calculation of w
    w = w_dep(
        v_bias=v_bias,
        epsilon=epsilon,
        n_acceptor=n_acceptor,
        n_donor=n_donor,
        x_cd=x_cd,
        temperature=temperature,
    )

    ao = np.pi * epsilon * eps0 * ((phi_implant / 2.0) ** 2 + d_implant * phi_implant)
    bo = np.pi * epsilon * eps0 * 2.0 * (d_implant + phi_implant)
    co = 3.0 * np.pi * epsilon * eps0
    return ao * 1.0 / w + bo + co * w


@njit
def dv_dt_cylindrical(
    v_bias: np.ndarray,
    photonic_current: np.ndarray,
    fixed_capacitance: float,
    sat_current: float,
    n: float,
    phi_implant: float,
    d_implant: float,
    n_acceptor: float,
    n_donor: float,
    x_cd: float,
    temperature: float,
) -> np.ndarray:
    """Use this diode discharge equation when considering a cylindrical diode.

    Parameters
    ----------
    v_bias : float,
        Diode polarisation
    photonic_current : float
        Photonic current
    fixed_capacitance : float
        Fixed capacitance
    sat_current : float
        Saturating current under dark conditions
    n : float
        Ideality factor
    phi_implant : float
        diameter of the implantation
    d_implant : float
        depth of the implantation
    n_acceptor : float
        density of acceptors
    n_donor : float
        density of donor
    x_cd : float
        cadmium concentrations
    temperature : float
        temperature of the sample

    Returns
    -------
    float
        Polarisation evolution per second.
    """
    const = Constants()
    e, k = const.electron_charge, const.k_b

    # Capacitance calculation as a function of Bias
    c = capa_pn_junction_cylindrical(
        v_bias=v_bias,
        phi_implant=phi_implant,
        d_implant=d_implant,
        n_acceptor=n_acceptor,
        n_donor=n_donor,
        x_cd=x_cd,
        temperature=temperature,
    )

    return -(
        e * sat_current * (np.exp(e * v_bias / (n * k * temperature)) - 1)
        - e * photonic_current
    ) / (fixed_capacitance + c)


@njit
def euler(
    time_step: float,
    nb_pts: int,
    v_bias: np.ndarray,
    phi_implant: float,
    d_implant: float,
    n_acceptor: float,
    n_donor: float,
    x_cd: float,
    temperature: float,
    photonic_current: np.ndarray,
    fixed_capacitance: float,
    sat_current: float,
    n: float,
):
    """Apply Euler method to solve the differential equation for detector non-linearity.

    Parameters
    ----------
    time_step : float
        Time step.
    nb_pts : float
        Number of points.
    v_bias : float
        Diode polarisation.
    phi_implant : float
        Diameter of the implantation.
    d_implant : float
        Depth of the implantation.
    n_acceptor : float
        Density of acceptors.
    n_donor : float
        Density of donor.
    x_cd : float
        Cadmium concentration.
    temperature : float
        Temperature of the sample.
    photonic_current : float
        Photonic current.
    fixed_capacitance : float
        Fixed capacitance.
    sat_current : float
        Saturating current under dark conditions.
    n : float
        Ideality factor.

    Returns
    -------
    ndarray:
        Voltage at the gate of the pixel SFD.
    """
    # Calculating step size
    h = time_step / nb_pts
    # Initialization of variables
    v = [v_bias]
    # ==============================
    #           SOLVE THE DIFFERENTIAL EQUATION
    # ==============================
    # EULER Method to solve the differential equation
    for _ in range(1, nb_pts):
        slope = dv_dt_cylindrical(
            v_bias=v[-1],
            photonic_current=photonic_current,
            fixed_capacitance=fixed_capacitance,
            sat_current=sat_current,
            n=n,
            phi_implant=phi_implant,
            d_implant=d_implant,
            n_acceptor=n_acceptor,
            n_donor=n_donor,
            x_cd=x_cd,
            temperature=temperature,
        )
        # Calculate new bias
        yn = v[-1] + h * slope
        # Saved the old bias
        v.append(yn)

    # Return bias evolution, capacitance and time
    return v[-1]



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import NamedTuple, Optional, Union

import numpy as np
import pandas as pd
import xarray as xr

from pyxel import load_datacube, load_image

#  from pyxel.pipelines import Processor
__all__ = [
    "CalibrationResult",
    "CalibrationMode",
    "Island",
    "check_ranges",
    "list_to_slice",
    "read_data",
    "create_processor_data_array",
    "read_datacubes",
    "list_to_3d_slice",
    "FitRange2D",
    "FitRange3D",
    "to_fit_range",
    "check_fit_ranges",
]


class CalibrationResult(NamedTuple):
    """Result class for calibration class."""

    dataset: xr.Dataset
    processors: pd.DataFrame
    logs: pd.DataFrame
    filenames: Sequence


# TODO: 'SingleModel' is never used. Remove this class ?
class CalibrationMode(Enum):
    """Enumeration for different modes of Calibration."""

    Pipeline = "pipeline"
    SingleModel = "single_model"  # TODO: This is never used


class Island(Enum):
    """User defines num_islands provides by Pygmo."""

    MultiProcessing = "multiprocessing"
    MultiThreading = "multithreading"
    IPyParallel = "ipyparallel"


def read_single_data(filename: Path) -> np.ndarray:
    """Read a numpy array from a FITS or NPY file."""
    # Late import to avoid circular import

    data = load_image(filename)

    # # TODO: Is it the right manner ?
    # if data is None:
    #     raise OSError(f"Input file '{filename}' can not be read by Pyxel.")

    return data


def read_single_datacube(filename: Path) -> np.ndarray:
    """Read a numpy data cube array from a FITS or NPY file."""

    data = load_datacube(filename)

    # # TODO: Is it the right manner ?
    # if data is None:
    #     raise OSError(f"Input file '{filename}' can not be read by Pyxel.")

    return data


def read_datacubes(filenames: Sequence[Path]) -> Sequence[np.ndarray]:
    """Read numpy array(s) from several FITS or NPY files."""
    output: Sequence[np.ndarray] = [
        read_single_datacube(Path(filename)) for filename in filenames
    ]

    return output


def read_data(filenames: Sequence[Path]) -> Sequence[np.ndarray]:
    """Read numpy array(s) from several FITS or NPY files."""
    output: Sequence[np.ndarray] = [
        read_single_data(Path(filename)) for filename in filenames
    ]

    return output


def create_processor_data_array(filenames: Sequence[Path]) -> xr.DataArray:
    """Create a 3D data array from FITS or NPY files for several processor(s).

    This DataArray will have dimensions 'processor', 'y' and 'x'

    Examples
    --------
    >>> create_processor_data_array(filenames=...)
    <xarray.DataArray (processor: 10, y: 2300, x: 1)>
    array(...)
    Coordinates:
      * processor  (processor) int64 0 1 2 3 4 5 6 7 8 9
      * y          (y) int64 0 1 2 3 4 5 6 7 ... 2293 2294 2295 2296 2297 2298 2299
      * x          (x) int64 0
    """
    data_3d = np.array([read_single_data(Path(filename)) for filename in filenames])

    num_processors, num_y, num_x = data_3d.shape

    data_array = xr.DataArray(
        data_3d,
        dims=["processor", "y", "x"],
        coords={
            "processor": range(num_processors),
            "y": range(num_y),
            "x": range(num_x),
        },
    )
    return data_array


# TODO: Refactor and add more unit tests. See #327
def list_to_slice(
    input_list: Optional[Sequence[int]] = None,
) -> Union[tuple[slice, slice], tuple[slice, slice, slice]]:
    if not input_list:
        return slice(None), slice(None)

    elif len(input_list) == 4:
        return slice(input_list[0], input_list[1]), slice(input_list[2], input_list[3])

    elif len(input_list) == 6:
        return (
            slice(input_list[0], input_list[1]),
            slice(input_list[2], input_list[3]),
            slice(input_list[4], input_list[5]),
        )

    else:
        raise ValueError("Fitting range should have 4 or 6 values")


# TODO: Refactor and add more unit tests. See #327
def list_to_3d_slice(
    input_list: Optional[Sequence[int]] = None,
) -> tuple[slice, slice, slice]:
    """TBW.

    :return:
    """
    if not input_list:
        return slice(None), slice(None), slice(None)

    elif len(input_list) == 6:
        return (
            slice(input_list[0], input_list[1]),
            slice(input_list[2], input_list[3]),
            slice(input_list[4], input_list[5]),
        )

    else:
        raise ValueError("Fitting range should have 6 values")


def slice_to_range(data: slice) -> range:
    """Convert a slice to a range.

    Examples
    --------
    >>> slice_to_range(slice(10))
    range(10)
    >>> slice_to_range(slice(10, 20))
    range(10, 20)
    >>> slice_to_range(slice(10, 20, 2))
    ValueError("Cannot use parameter 'step' in the slice object.")

    Raises
    ------
    ValueError
        If the input slice is not correct.
    """
    if data.step is not None:
        raise ValueError("Cannot use parameter 'step' in the slice object.")
    if data.stop is None:
        raise ValueError("Missing 'stop' parameter in the slice object")

    if data.start is None:
        result = range(data.stop)

        if data.stop <= 0:
            raise ValueError("Parameter 'stop' must be strictly greater than 'start'")
    else:
        if data.start < 0:
            raise ValueError("Parameter 'start' must be strictly positive")
        if data.start >= data.stop:
            raise ValueError("Parameter 'stop' must be strictly greater than 'start'")

        result = range(data.start, data.stop)

    return result


def _check_out_ranges(
    target_fit_range: Sequence[int],
    out_fit_range: Sequence[int],
) -> None:
    if len(out_fit_range) not in (4, 6):
        raise ValueError

    if (target_fit_range[1] - target_fit_range[0]) != (
        out_fit_range[1] - out_fit_range[0]
    ):
        raise ValueError("Fitting ranges have different lengths in 1st dimension")

    # TODO: Refactor and add more unit tests. See #328
    if len(target_fit_range) == len(out_fit_range) == 4:
        if (target_fit_range[3] - target_fit_range[2]) != (
            out_fit_range[3] - out_fit_range[2]
        ):
            raise ValueError("Fitting ranges have different lengths in 2nd dimension")

    if len(target_fit_range) == len(out_fit_range) == 6:
        if (target_fit_range[5] - target_fit_range[4]) != (
            out_fit_range[5] - out_fit_range[4]
        ):
            raise ValueError("Fitting ranges have different lengths in third dimension")


# TODO: Refactor and add more unit tests. See #328
def check_ranges(
    target_fit_range: Sequence[int],
    out_fit_range: Sequence[int],
    rows: int,
    cols: int,
    readout_times: Optional[int] = None,
) -> None:
    """TBW."""
    if not target_fit_range:
        return

    if len(target_fit_range) not in (4, 6):
        raise ValueError

    if out_fit_range:
        _check_out_ranges(
            target_fit_range=target_fit_range, out_fit_range=out_fit_range
        )

    # for i in [0, 1]:
    #     # TODO: Refactor and add more unit tests. See #328
    #     if not (0 <= target_fit_range[i] <= rows):
    #         raise ValueError("Value of target fit range is wrong")

    for i in (-3, -4):
        if not (0 <= target_fit_range[i] <= rows):
            raise ValueError("Value of target fit range is wrong")

    for i in (-1, -2):
        if not (0 <= target_fit_range[i] <= cols):
            raise ValueError("Value of target fit range is wrong")

    if len(target_fit_range) == 6:
        if readout_times is None:
            raise ValueError("Target data is not a 3 dimensional array")

        for i in (0, 1):
            # TODO: Refactor and add more unit tests. See #328
            if not (0 <= target_fit_range[i] <= readout_times):
                raise ValueError("Value of target fit range is wrong")


@dataclass(frozen=True)
class FitRange2D:
    """Represent a 2D range or slice with a row range and a column range.

    Parameters
    ----------
    row : slice
        Range of rows in the 2D range
    col : slice
        Range of columns in the 2D range

    Examples
    --------
    >>> range2d = FitRange2D(row=slice(0, 5), col=slice(2, 7))
    >>> row_slice, col_slice = range2d.to_slices()
    >>> row_slice
    slice(0, 5)
    >>> col_slice
    slice(2,7)

    >>> FitRange2D.from_sequence([0, 5, 2, 7])
    FitRange2D(slice(0, 5),slice(2,7))
    """

    row: slice
    col: slice

    @classmethod
    def from_sequence(cls, data: Sequence[Optional[int]]) -> "FitRange2D":
        if not data:
            data = [None] * 4

        if len(data) != 4:
            raise ValueError("Fitting range should have 4 values")

        y_start, y_stop, x_start, x_stop = data
        return cls(row=slice(y_start, y_stop), col=slice(x_start, x_stop))

    def to_dict(self) -> Mapping[str, slice]:
        return {"y": self.row, "x": self.col}

    def to_slices(self) -> tuple[slice, slice]:
        return self.row, self.col

    def check(self, rows: int, cols: int):
        if not self.row.stop <= rows:
            raise ValueError("Value of target fit range is wrong")

        if not self.col.stop <= cols:
            raise ValueError("Value of target fit range is wrong")


@dataclass(frozen=True)
class FitRange3D:
    """Represent a 3D range or slice with a time range, row range and a column range.

    Parameters
    ----------
    time : FitSlice
         Range of time in the 3D range
    row : FitSlice
        Range of rows in the 3D range
    col : FitSlice
        Range of columns in the 3D range

    Examples
    --------
    >>> range3d = FitRange3D(time=slice(0, 10), row=slice(0, 5), col=slice(2, 7))
    >>> time_slice, row_slice, col_slice = range3d.to_slices()
    >>> time_slice
    slice(0, 10)
    >>> row_slice
    slice(0, 5)
    >>> col_slice
    slice(2,7)

    >>> FitRange3D.from_sequence([0, 10, 0, 5, 2, 7])
    FitRange3D(time=slice(0,10), row=slice(0, 5), col=slice(2,7))
    """

    time: slice
    row: slice
    col: slice

    @classmethod
    def from_sequence(cls, data: Sequence[Optional[int]]) -> "FitRange3D":
        if not data:
            data = [None] * 6

        if len(data) == 4:
            data = [None, None, *data]

        if len(data) != 6:
            raise ValueError("Fitting range should have 6 values")

        time_start, time_stop, y_start, y_stop, x_start, x_stop = data
        return cls(
            time=slice(time_start, time_stop),
            row=slice(y_start, y_stop),
            col=slice(x_start, x_stop),
        )

    def to_dict(self) -> Mapping[str, slice]:
        return {"time": self.time, "y": self.row, "x": self.col}

    def to_slices(self) -> tuple[slice, slice, slice]:
        return self.time, self.row, self.col

    def check(self, rows: int, cols: int, readout_times: Optional[int] = None):
        if not self.row.stop <= rows:
            raise ValueError("Value of target fit range is wrong")

        if not self.col.stop <= cols:
            raise ValueError("Value of target fit range is wrong")

        if readout_times is None:
            raise ValueError("Target data is not a 3 dimensional array")

        if not self.time.stop <= readout_times:
            raise ValueError("Value of target fit range is wrong")


def to_fit_range(
    input_list: Optional[Sequence[int]] = None,
) -> Union[FitRange2D, FitRange3D]:
    if not input_list:
        return FitRange2D(row=slice(None), col=slice(None))

    elif len(input_list) == 4:
        return FitRange2D.from_sequence(input_list)

    elif len(input_list) == 6:
        return FitRange3D.from_sequence(input_list)

    else:
        raise ValueError("Fitting range should have 4 or 6 values")


def _check_out_fit_ranges(
    target_fit_range: Union[FitRange2D, FitRange3D],
    out_fit_range: Union[FitRange2D, FitRange3D],
):
    if (
        isinstance(target_fit_range, FitRange3D)
        and isinstance(out_fit_range, FitRange3D)
        and target_fit_range.time.stop != out_fit_range.time.stop
    ):
        raise ValueError(
            "Fitting ranges have different lengths in dimension 'readout time'"
        )

    if target_fit_range.row.stop != out_fit_range.row.stop:
        raise ValueError("Fitting ranges have different lengths in dimension 'y'")

    if target_fit_range.col.stop != out_fit_range.col.stop:
        raise ValueError("Fitting ranges have different lengths in dimension 'x'")


# TODO: Refactor and add more unit tests. See #328
def check_fit_ranges(
    target_fit_range: Union[FitRange2D, FitRange3D, None],
    out_fit_range: Union[FitRange2D, FitRange3D, None],
    rows: int,
    cols: int,
    readout_times: Optional[int] = None,
) -> None:
    """Check if ``target_fit_range`` and ``out_fit_range`` are valid.

    This functions checks if ``target_fit_range`` is valid for the specified ``rows`` and ``columns``
    and if ``out_fit_range`` is compatible with ``target_fit_range``.

    Parameters
    ----------
    target_fit_range : FitRange2D, FitRange3D. Optional
        A target range to check if it's valid or not.
    out_fit_range : FitRange2D, FitRange3D. Optional
        An output range to check whether it's compatible with the target fit range.
    rows : int
        Number of rows.
    cols : int
        Number of columns
    readout_times : int, Optional
        Number of readout times. This parameter is only used if the target fit range is
        a 3D range.

    Raises
    ------
    ValueError
        If ``target_fit_range`` is not valid or not compatible with ``out_fit_range``.
    """
    if not target_fit_range:
        return

    if out_fit_range:
        _check_out_fit_ranges(
            target_fit_range=target_fit_range, out_fit_range=out_fit_range
        )

    if isinstance(target_fit_range, FitRange2D):
        target_fit_range.check(rows=rows, cols=cols)
    else:
        target_fit_range.check(rows=rows, cols=cols, readout_times=readout_times)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Fix pattern noise model."""

from pathlib import Path
from typing import Literal, Optional, Union

import numpy as np

from pyxel.detectors import Detector, Geometry
from pyxel.util import load_cropped_and_aligned_image, set_random_seed


def fpn_from_file(
    geometry: Geometry,
    filename: Union[Path, str],
    position: tuple[int, int] = (0, 0),
    align: Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ] = None,
) -> np.ndarray:
    """Get fixed pattern noise caused by pixel non-uniformity during charge collection from a file.

    Parameters
    ----------
    geometry : Geometry
        Geometry of detector.
    filename : str or Path
        Path to a file with an array or an image.
    position : tuple
        Indices of starting row and column, used when fitting noise to detector.
    align : Literal
        Keyword to align the noise to detector. Can be any from:
        ("center", "top_left", "top_right", "bottom_left", "bottom_right")

    Raises
    ------
    FileNotFoundError
        If the folder path is not found.

    Returns
    -------
    prnu_2d : ndarray
        Fixed pattern noise caused by pixel non-uniformity during charge collection.
    """
    position_y, position_x = position

    folder_path = Path(filename).expanduser().resolve()

    if not folder_path.exists():
        raise FileNotFoundError(f"Cannot find folder '{folder_path}' !")

    # Load charge profile as numpy array.
    prnu_2d: np.ndarray = load_cropped_and_aligned_image(
        shape=(geometry.row, geometry.col),
        filename=filename,
        position_x=position_x,
        position_y=position_y,
        align=align,
    )

    return prnu_2d


def compute_simple_prnu(
    shape: tuple[int, int],
    quantum_efficiency: float,
    fixed_pattern_noise_factor: float,
) -> np.ndarray:
    """Compute fixed pattern noise caused by pixel non-uniformity during charge collection.

    Parameters
    ----------
    shape : tuple
        Output array shape.
    quantum_efficiency : float
        Quantum efficiency of detector.
    fixed_pattern_noise_factor : float
        Fixed pattern noise factor.

    Returns
    -------
    prnu_2d : ndarray
        Fixed pattern noise caused by pixel non-uniformity during charge collection.
    """

    prnu_2d = np.ones(shape) * quantum_efficiency
    prnu_sigma = quantum_efficiency * fixed_pattern_noise_factor
    prnu_2 = prnu_2d * (1 + np.random.lognormal(sigma=prnu_sigma, size=shape))

    return prnu_2


def fixed_pattern_noise(
    detector: Detector,
    filename: Union[Path, str, None] = None,
    position: tuple[int, int] = (0, 0),
    align: Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ] = None,
    fixed_pattern_noise_factor: Optional[float] = None,
    seed: Optional[int] = None,
) -> None:
    """Add fixed pattern noise caused by pixel non-uniformity during charge collection.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    filename : str or Path or None
        Path to a file with an array or an image.
    position : tuple
        Indices of starting row and column, used when fitting noise to detector.
    align : Literal
        Keyword to align the noise to detector. Can be any from:
        ("center", "top_left", "top_right", "bottom_left", "bottom_right")
    fixed_pattern_noise_factor : float, optional
        Fixed pattern noise factor.
    seed : int, optional
        Random seed.

    Raises
    ------
    ValueError
        If no filename and no fixed_pattern_noise_factor is giving or both are giving.
    """

    geo = detector.geometry  # type: Geometry
    char = detector.characteristics

    with set_random_seed(seed):
        if filename is not None and fixed_pattern_noise_factor is None:
            prnu_2d = fpn_from_file(
                geometry=geo,
                filename=filename,
                position=position,
                align=align,
            )

        elif fixed_pattern_noise_factor is not None and filename is None:
            prnu_2d = compute_simple_prnu(
                shape=geo.shape,
                quantum_efficiency=char.quantum_efficiency,
                fixed_pattern_noise_factor=fixed_pattern_noise_factor,
            )

        else:
            raise ValueError(
                "Either filename or fixed_pattern_noise_factor has to be defined."
            )
    detector.pixel.array *= prnu_2d



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""The detector attribute and data structure subpackage."""

# flake8: noqa
# Warning: Import order matters
from .readout_properties import ReadoutProperties
from .environment import Environment
from .detector import Detector
from .geometry import Geometry
from .characteristics import Characteristics
from .ccd import CCDGeometry, CCD
from .mkid import MKID, MKIDGeometry
from .apd import APD, APDCharacteristics, APDGeometry
from .cmos import CMOSGeometry, CMOS
from .optics import Optics



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import inspect
import warnings
from collections.abc import Mapping, MutableMapping
from typing import TYPE_CHECKING, Any, Callable, Optional

from pyxel.evaluator import evaluate_reference

if TYPE_CHECKING:
    import numpy as np

    from pyxel.calibration import FittingCallable
    from pyxel.detectors import Detector


class Arguments(MutableMapping):
    """Arguments class for usage in ModelFunction.

    Class Arguments is initialized from a dictionary of model function arguments.
    It resembles the behaviour of the passed dictionary with additional methods __setattr__ and __getattr__,
    which enable to get and set the model parameters through attribute interface.
    Dictionary of arguments is saved in private attribute _arguments.

    Examples
    --------
    >>> from pyxel.pipelines.model_function import Arguments
    >>> arguments = Arguments({"one": 1, "two": 2})
    >>> arguments
    Arguments({'one': 1, 'two': 2})

    Access arguments
    >>> arguments["one"]
    1
    or
    >>> arguments.one
    1

    Changing parameters
    >>> arguments["one"] = 10
    >>> arguments["two"] = 20
    >>> arguments
    Arguments({'one': 10, 'two': 20})

    Non existing arguments
    >>> arguments["three"] = 3
    KeyError: 'No argument named three !'
    >>> arguments.three = 3
    AttributeError: 'No argument named three !'
    """

    def __init__(self, input_arguments: dict[str, Any]):
        self._arguments: dict[str, Any] = input_arguments

    def __setitem__(self, key, value):
        if key not in self._arguments:
            raise KeyError(f"No argument named {key} !")

        self._arguments[key] = value

    def __getitem__(self, key):
        if key not in self._arguments:
            raise KeyError(f"No argument named {key} !")

        return self._arguments[key]

    def __delitem__(self, key):
        del self._arguments[key]

    def __iter__(self):
        return iter(self._arguments)

    def __len__(self):
        return len(self._arguments)

    def __getattr__(self, key):
        # Use non-modified __getattr__ in this case.
        if key == "_arguments":
            return object.__getattribute__(self, "_arguments")

        if key not in self._arguments:
            raise AttributeError(f"'{self!r}' object has no argument {key!r} !")

        return self._arguments[key]

    def __setattr__(self, key, value):
        # Use non-modified __setattr__ in this case.
        if key == "_arguments":
            super().__setattr__(key, value)
            return

        if key not in self._arguments:
            raise AttributeError(f"'{self!r}' object has no argument {key!r} !")

        self._arguments[key] = value

    def __dir__(self):
        return dir(type(self)) + list(self)

    def __repr__(self):
        return f"Arguments({self._arguments})"

    # def __deepcopy__(self, memo) -> "Arguments":
    #     """TBW."""
    #     return Arguments(deepcopy(self._arguments))


# TODO: Improve this class. See issue #132.
# TODO: Add unit tests #182
class ModelFunction:
    """Create a wrapper function around a Model function.

    Parameters
    ----------
    func : str
        Full name of the model to use with this ModelFunction.
    name : str
        A unique name for this ModelFunction.
    arguments : dict (optional). Default: None
        Parameters that will be passed to 'func'.
    enabled : bool. Default: True
        A flag indicating whether this ModelFunction is enabled or disabled.

    Examples
    --------
    >>> model_func = ModelFunction(
    ...     func="pyxel.models.photon_collection.illumination",
    ...     name="illumination",
    ...     arguments={"level": 1, "option": "foo"},
    ... )

    Access basic parameters

    >>> model_func.name
    'illumination'
    >>> model_func.enabled
    True
    >>> model_func.arguments
    Arguments({'level': 1, 'option': 'foo'})

    Access the arguments with a ``dict`` interface

    >>> list(model_func.arguments)
    ['level', 'option']
    >>> model_func.arguments["level"]
    1
    >>> model_func.arguments["level"] = 2
    TypeError: 'Arguments' object does not support item assignment

    Access the arguments with an attribute interface

    >>> model_func.arguments.level
    1

    Execute this ModelFunction
    >>> from pyxel.detectors import CCD
    >>> detector = CCD(...)
    >>> model_func(detector=detector)
    >>> detector  # Modified detector
    """

    def __init__(
        self,
        func: str,
        name: str,
        arguments: Optional[dict] = None,
        enabled: bool = True,
    ):
        if inspect.isclass(func):
            raise AttributeError("Cannot pass a class to ModelFunction.")

        self._func_name: str = func
        self._func: Optional[Callable] = None

        self._name: str = name
        self.enabled: bool = enabled

        if arguments is None:
            arguments = {}

        self._arguments = Arguments(arguments)

        # self.group = None               # TODO

    def __repr__(self) -> str:
        cls_name: str = self.__class__.__name__

        return (
            f"{cls_name}(name={self.name!r}, func={self._func}, "
            f"arguments={self.arguments!r}, enabled={self.enabled!r})"
        )

    @property
    def name(self) -> str:
        """Name for this ModelFunction."""
        return self._name

    @property
    def arguments(self) -> Arguments:
        """TBW."""
        return self._arguments

    @property
    def func(self) -> Callable:
        """TBW."""
        if self._func is None:
            self._func = evaluate_reference(self._func_name)

        if hasattr(self._func, "__deprecated__"):
            # This will be visible by the user
            msg = self._func.__deprecated__
            warnings.warn(msg, FutureWarning, stacklevel=2)

        return self._func

    def __call__(self, detector: "Detector") -> None:
        """TBW."""
        self.func(detector, **self.arguments)


class FitnessFunction:
    """Fitness function for model fitting."""

    def __init__(self, func: str, arguments: Optional[Mapping[str, Any]] = None):
        self._func: FittingCallable = evaluate_reference(func)
        self._arguments: Optional[Mapping[str, Any]] = arguments

    def __call__(
        self,
        simulated: "np.ndarray",
        target: "np.ndarray",
        weighting: "np.ndarray",
    ) -> float:
        """Compute fitness."""
        if self._arguments is None:
            result: float = self._func(
                simulated=simulated,
                target=target,
                weighting=weighting,
            )
        else:
            result = self._func(
                simulated=simulated,
                target=target,
                weighting=weighting,
                **self._arguments,
            )
        return result



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


"""Voltage offset model tests."""
import pytest

from pyxel.detectors import (
    APD,
    CCD,
    APDCharacteristics,
    APDGeometry,
    CCDGeometry,
    Characteristics,
    Environment,
)
from pyxel.models.charge_measurement import dc_offset, output_pixel_reset_voltage_apd


@pytest.fixture
def ccd_5x5() -> CCD:
    """Create a valid CCD detector."""
    return CCD(
        geometry=CCDGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(
            adc_voltage_range=(0.0, 10.0),
        ),
    )


@pytest.fixture
def apd_5x5() -> APD:
    """Create a valid CCD detector."""
    return APD(
        geometry=APDGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=APDCharacteristics(
            quantum_efficiency=1.0,
            adc_voltage_range=(0.0, 10.0),
            adc_bit_resolution=16,
            full_well_capacity=100000,
            avalanche_gain=1.0,
            pixel_reset_voltage=5.0,
            roic_gain=0.8,
        ),
    )


@pytest.mark.parametrize("detector_type", ["ccd", "apd"])
@pytest.mark.parametrize(
    "voltage",
    [pytest.param(0.1, id="ccd"), pytest.param(3.0, id="apd")],
)
def test_offset(
    ccd_5x5: CCD,
    apd_5x5: APD,
    detector_type: str,
    voltage: float,
):
    """Test offset models 'dc_offset' and 'output_pixel_reset_voltage_apd' with valid inputs."""

    if detector_type == "ccd":
        dc_offset(detector=ccd_5x5, offset=voltage)
    elif detector_type == "apd":
        output_pixel_reset_voltage_apd(detector=apd_5x5, roic_drop=voltage)
    else:
        raise NotImplementedError


@pytest.mark.parametrize(
    "voltage, exp_exc, exp_error",
    [
        pytest.param(
            20.0,
            ValueError,
            "Parameter 'offset' out of bonds of the ADC voltage range.",
        ),
        pytest.param(
            -1.0,
            ValueError,
            "Parameter 'offset' out of bonds of the ADC voltage range.",
        ),
    ],
)
def test_dc_offset_invalid(ccd_5x5: CCD, voltage: float, exp_exc, exp_error):
    """Test model 'dc_offset' with valid inputs."""
    detector = ccd_5x5
    with pytest.raises(exp_exc, match=exp_error):
        dc_offset(detector=detector, offset=voltage)


@pytest.mark.parametrize(
    "drop, exp_exc, exp_error",
    [
        pytest.param(
            10.0,
            ValueError,
            "Output pixel reset voltage out of bonds of the ADC voltage range.",
        ),
    ],
)
def test_output_pixel_reset_voltage_apd_invalid(
    apd_5x5: APD, drop: float, exp_exc, exp_error
):
    """Test model 'output_pixel_reset_voltage_apd' with valid inputs."""
    detector = apd_5x5
    with pytest.raises(exp_exc, match=exp_error):
        output_pixel_reset_voltage_apd(detector=detector, roic_drop=drop)


def test_output_pixel_reset_voltage_apd_invalid_with_ccd(ccd_5x5: CCD):
    """Test model 'output_pixel_reset_voltage_apd' with a 'CCD'."""
    detector = ccd_5x5

    with pytest.raises(TypeError, match="Expecting a 'APD' detector object."):
        output_pixel_reset_voltage_apd(detector=detector, roic_drop=1.0)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


from collections.abc import Sequence

import numpy as np
import pytest

from pyxel.detectors import (
    CCD,
    CMOS,
    CCDGeometry,
    Characteristics,
    CMOSGeometry,
    Environment,
    ReadoutProperties,
)
from pyxel.models.charge_measurement import (
    output_node_linearity_poly,
    physical_non_linearity,
    physical_non_linearity_with_saturation,
    simple_physical_non_linearity,
)


@pytest.fixture
def ccd_5x5() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(temperature=80),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


@pytest.fixture
def cmos_5x5() -> CMOS:
    """Create a valid CMOS detector."""
    detector = CMOS(
        geometry=CMOSGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(temperature=80),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


@pytest.mark.parametrize(
    "coefficients",
    [
        pytest.param([0, 1, 0.9]),
        pytest.param([5, 0.5, 0.9, 0.8]),
        pytest.param([3]),
        pytest.param([0, 3]),
    ],
)
def test_non_linearity_valid(ccd_5x5: CCD, coefficients: Sequence):
    """Test model 'non_linearity' with valid inputs."""
    detector = ccd_5x5
    detector.signal.array = np.ones(detector.signal.shape)
    output_node_linearity_poly(detector=detector, coefficients=coefficients)


@pytest.mark.parametrize(
    "coefficients, exp_exc, exp_error",
    [
        pytest.param(
            [], ValueError, "Length of coefficient list should be more than 0."
        ),
        pytest.param(
            [0, -1],
            ValueError,
            "Signal array contains negative values after applying non-linearity model!",
        ),
    ],
)
def test_non_linearity_invalid(
    ccd_5x5: CCD, coefficients: Sequence, exp_exc, exp_error
):
    """Test model 'non_linearity' with valid inputs."""
    detector = ccd_5x5
    detector.signal.array = np.ones(detector.signal.shape)
    with pytest.raises(exp_exc, match=exp_error):
        output_node_linearity_poly(detector=detector, coefficients=coefficients)


def test_simple_physical_non_linearity_valid(cmos_5x5: CMOS):
    """Test model 'simple_physical_non_linearity' with valid inputs."""
    detector = cmos_5x5
    detector.signal.array = np.ones(detector.signal.shape)
    simple_physical_non_linearity(
        detector=detector,
        cutoff=2.0,
        n_acceptor=1.0e18,
        n_donor=1.0e15,
        diode_diameter=5.0,
        v_bias=0.1,
    )


def test_physical_non_linearity_valid(cmos_5x5: CMOS):
    """Test model 'physical_non_linearity' with valid inputs."""
    detector = cmos_5x5
    detector.signal.array = np.ones(detector.signal.shape)
    physical_non_linearity(
        detector=detector,
        cutoff=2.0,
        n_acceptor=1.0e18,
        n_donor=1.0e15,
        diode_diameter=5.0,
        v_bias=0.1,
        fixed_capacitance=5.0e-15,
    )


def test_physical_non_linearity_with_saturation_valid(cmos_5x5: CMOS):
    """Test model 'physical_non_linearity_with_saturation' with valid inputs."""
    detector = cmos_5x5
    detector.signal.array = np.ones(detector.signal.shape)
    physical_non_linearity_with_saturation(
        detector=detector,
        cutoff=2.0,
        n_acceptor=1.0e18,
        n_donor=1.0e15,
        phi_implant=5.0,
        d_implant=2.0,
        saturation_current=0.001,
        ideality_factor=1.34,
        v_reset=0.0,
        d_sub=0.220,
        fixed_capacitance=5.0e-15,
        euler_points=100,
    )


def test_simple_physical_non_linearity_with_ccd(ccd_5x5: CCD):
    """Test model 'simple_physical_non_linearity' with a 'CCD'."""
    detector = ccd_5x5

    with pytest.raises(TypeError, match="Expecting a 'CMOS' detector object."):
        simple_physical_non_linearity(
            detector=detector,
            cutoff=2.0,
            n_acceptor=1.0e18,
            n_donor=1.0e15,
            diode_diameter=5.0,
            v_bias=0.1,
        )


def test_physical_non_linearity_with_ccd(ccd_5x5: CCD):
    """Test model 'physical_non_linearity' with a 'CCD'."""
    detector = ccd_5x5

    with pytest.raises(TypeError, match="Expecting a 'CMOS' detector object."):
        physical_non_linearity(
            detector=detector,
            cutoff=2.0,
            n_acceptor=1.0e18,
            n_donor=1.0e15,
            diode_diameter=5.0,
            v_bias=0.1,
            fixed_capacitance=5.0e-15,
        )


def test_physical_non_linearity_with_saturation_with_ccd(ccd_5x5: CCD):
    """Test model 'physical_non_linearity_with_saturation' with a 'CCD'."""
    detector = ccd_5x5

    with pytest.raises(TypeError, match="Expecting a 'CMOS' detector object."):
        physical_non_linearity_with_saturation(
            detector=detector,
            cutoff=2.0,
            n_acceptor=1.0e18,
            n_donor=1.0e15,
            phi_implant=5.0,
            d_implant=2.0,
            saturation_current=0.001,
            ideality_factor=1.34,
            v_reset=0.0,
            d_sub=0.220,
            fixed_capacitance=5.0e-15,
            euler_points=100,
        )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import sys
import typing as t

import click
import requests


@click.command()
@click.option("--environment", default="production", type=str, show_default=True)
@click.option(
    "--ci_api_v4_url",
    envvar="CI_API_V4_URL",
    type=str,
    help="The GitLab API v4 root URL.",
    show_default=True,
)
@click.option(
    "--ci_project_id",
    envvar="CI_PROJECT_ID",
    type=int,
    help="The project's ID",
    required=True,
)
@click.option(
    "--private_token",
    envvar="PRIVATE_TOKEN",
    required=False,
    help="A token to authenticate",
)
def main(
    environment: str,
    ci_api_v4_url: str,
    ci_project_id: int,
    private_token: t.Optional[str],
):
    # Create token
    headers = {}

    if private_token:
        headers["PRIVATE-TOKEN"] = private_token
    else:
        raise ValueError("Missing 'private_token'.")

    # Get main http address
    addr = f"{ci_api_v4_url}/projects/{ci_project_id}"

    # Get list of environments
    # GET /projects/:id/environments
    url_env_id = f"{addr}/environments?name={environment}"

    r = requests.get(url_env_id, headers=headers)
    r.raise_for_status()

    rsp = r.json()
    environment_id: int = rsp[0]["id"]

    # Get a specific environment
    # GET /projects/:id/environments/:environment_id
    url_env = f"{addr}/environments/{environment_id}"

    r = requests.get(url_env, headers=headers)
    r.raise_for_status()

    rsp = r.json()
    deployable_dct: t.Mapping = rsp["last_deployment"]["deployable"]
    job_id: int = deployable_dct["id"]
    # artifact_filename = deployable_dct["artifacts_file"]["filename"]  # type: str
    artifact_filename: int = "artifacts.zip"

    # Download artifact
    url_download = f"{addr}/jobs/{job_id}/artifacts"

    r = requests.get(url_download, headers=headers)
    r.raise_for_status()

    # Save artifact
    with open(artifact_filename, "wb") as fh:
        fh.write(r.content)

    print(f"File {artifact_filename!r} is successfully downloaded !")
    sys.exit()


if __name__ == "__main__":
    main()



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW.

References
----------
[1] Leonardo MW Ltd., Electrical Interface Document for SAPHIRA (ME1000) in a 68-pin LLC
(3313-520-), issue 1A, 2012.

[2] I. Pastrana et al., HgCdTe SAPHIRA arrays: individual pixel measurement of charge gain and
node capacitance utilizing a stable IR LED, in High Energy, Optical, and Infrared Detectors for
Astronomy VIII, 2018, vol. 10709, no. July 2018, 2018, p. 37.

[3] S. B. Goebel et al., Overview of the SAPHIRA detector for adaptive optics applications, in
Journal of Astronomical Telescopes, Instruments, and Systems, 2018, vol. 4, no. 02, p. 1.

[4] G. Finger et al., Sub-electron read noise and millisecond full-frame readout with the near
infrared eAPD array SAPHIRA, in Adaptive Optics Systems V, 2016, vol. 9909, no. July 2016, p.
990912.

[5] I. M. Baker et al., Linear-mode avalanche photodiode arrays in HgCdTe at Leonardo, UK: the
current status, in Image Sensing Technologies: Materials, Devices, Systems, and Applications VI,
2019, vol. 10980, no. May, p. 20.
"""

import math
from collections.abc import Mapping
from typing import Optional

import numpy as np
from toolz import dicttoolz

from pyxel.util.memory import get_size


class APDCharacteristics:
    """Characteristic attributes of the APD detector.

    Parameters
    ----------
    roic_gain
        Gain of the read-out integrated circuit. Unit: V/V
    quantum_efficiency : float, optional
        Quantum efficiency.
    full_well_capacity : float, optional
        Full well capacity. Unit: e-
    adc_bit_resolution : int, optional
        ADC bit resolution.
    adc_voltage_range : tuple of floats, optional
        ADC voltage range. Unit: V
    avalanche_gain : float, optional
        APD gain. Unit: electron/electron
    pixel_reset_voltage : float
        DC voltage going into the detector, not the voltage of a reset pixel. Unit: V
    common_voltage : float
        Common voltage. Unit: V
    """

    def __init__(
        self,
        roic_gain: float,  # unit: V
        quantum_efficiency: Optional[float] = None,  # unit: NA
        full_well_capacity: Optional[float] = None,  # unit: electron
        adc_bit_resolution: Optional[int] = None,
        adc_voltage_range: Optional[tuple[float, float]] = None,  # unit: V
        avalanche_gain: Optional[float] = None,  # unit: electron/electron
        pixel_reset_voltage: Optional[float] = None,  # unit: V
        common_voltage: Optional[float] = None,  # unit: V
    ):
        self._avalanche_gain = avalanche_gain
        self._common_voltage = common_voltage
        self._pixel_reset_voltage = pixel_reset_voltage

        if avalanche_gain and pixel_reset_voltage and common_voltage:
            raise ValueError(
                "Please only specify two inputs out of: avalanche gain, pixel reset voltage, common voltage."
            )
        elif avalanche_gain and pixel_reset_voltage and not common_voltage:
            self._avalanche_bias = self.gain_to_bias_saphira(avalanche_gain)
            self._common_voltage = pixel_reset_voltage - self.avalanche_bias
        elif avalanche_gain and common_voltage and not pixel_reset_voltage:
            self._avalanche_bias = self.gain_to_bias_saphira(avalanche_gain)
            self._pixel_reset_voltage = common_voltage + self.avalanche_bias
        elif common_voltage and pixel_reset_voltage and not avalanche_gain:
            self._avalanche_bias = pixel_reset_voltage - common_voltage
            self._avalanche_gain = self.bias_to_gain_saphira(self.avalanche_bias)
        else:
            raise ValueError(
                "Not enough input parameters provided to calculate avalanche bias!"
            )

        if quantum_efficiency and not (0.0 <= quantum_efficiency <= 1.0):
            raise ValueError("'quantum_efficiency' must be between 0.0 and 1.0.")
        if avalanche_gain and not (1.0 <= avalanche_gain <= 1000.0):
            raise ValueError("'apd_gain' must be between 1.0 and 1000.0.")
        if adc_bit_resolution and not (4 <= adc_bit_resolution <= 64):
            raise ValueError("'adc_bit_resolution' must be between 4 and 64.")
        if adc_voltage_range and not len(adc_voltage_range) == 2:
            raise ValueError("Voltage range must have length of 2.")
        if full_well_capacity and not (0.0 <= full_well_capacity <= 1.0e7):
            raise ValueError("'full_well_capacity' must be between 0 and 1e7.")

        self._quantum_efficiency = quantum_efficiency
        self._full_well_capacity = full_well_capacity
        self._adc_voltage_range = adc_voltage_range
        self._adc_bit_resolution = adc_bit_resolution
        self._node_capacitance: float = self.bias_to_node_capacitance_saphira(
            self.avalanche_bias
        )
        self._roic_gain = roic_gain
        self._charge_to_volt_conversion: float = self.detector_gain_saphira(
            capacitance=self.node_capacitance, roic_gain=self.roic_gain
        )
        self._numbytes = 0

    def __eq__(self, other) -> bool:
        return (
            type(self) is type(other)
            and self._quantum_efficiency == other._quantum_efficiency
            and self._full_well_capacity == other._full_well_capacity
            and self._adc_bit_resolution == other._adc_bit_resolution
            and self._adc_voltage_range == other._adc_voltage_range
            and self._avalanche_gain == other._avalanche_gain
            and self._pixel_reset_voltage == other._pixel_reset_voltage
            and self._common_voltage == other._common_voltage
        )

    @property
    def quantum_efficiency(self) -> float:
        """Get Quantum efficiency."""
        if self._quantum_efficiency:
            return self._quantum_efficiency
        else:
            raise ValueError(
                "'quantum_efficiency' not specified in detector characteristics."
            )

    @quantum_efficiency.setter
    def quantum_efficiency(self, value: float) -> None:
        """Set Quantum efficiency."""
        if np.min(value) < 0.0 and np.max(value) <= 1.0:
            raise ValueError("'quantum_efficiency' values must be between 0.0 and 1.0.")

        self._quantum_efficiency = value

    @property
    def avalanche_gain(self) -> float:
        """Get APD gain."""
        if self._avalanche_gain:
            return self._avalanche_gain
        else:
            raise ValueError("'apd_gain' not specified in detector characteristics.")

    @avalanche_gain.setter
    def avalanche_gain(self, value: float) -> None:
        """Set APD gain."""
        if np.min(value) < 1.0 and np.max(value) <= 1000.0:
            raise ValueError("'apd_gain' values must be between 1.0 and 1000.")
        self._avalanche_gain = value
        self._avalanche_bias = self.gain_to_bias_saphira(value)
        self._common_voltage = self.pixel_reset_voltage - self.avalanche_bias

    @property
    def pixel_reset_voltage(self) -> float:
        """Get pixel reset voltage."""
        if self._pixel_reset_voltage:
            return self._pixel_reset_voltage
        else:
            raise ValueError("'apd_gain' not specified in detector characteristics.")

    @pixel_reset_voltage.setter
    def pixel_reset_voltage(self, value: float) -> None:
        """Set pixel reset voltage."""
        self._avalanche_bias = value - self.common_voltage
        self._avalanche_gain = self.bias_to_gain_saphira(self.avalanche_bias)
        self._pixel_reset_voltage = value

    @property
    def common_voltage(self) -> float:
        """Get common voltage."""
        if self._common_voltage:
            return self._common_voltage
        else:
            raise ValueError("'apd_gain' not specified in detector characteristics.")

    @common_voltage.setter
    def common_voltage(self, value: float) -> None:
        """Set common voltage."""
        self._avalanche_bias = self.pixel_reset_voltage - value
        self._avalanche_gain = self.bias_to_gain_saphira(self.avalanche_bias)
        self._common_voltage = value

    @property
    def avalanche_bias(self) -> float:
        """Get avalanche bias."""
        if self._avalanche_bias:
            return self._avalanche_bias
        else:
            raise ValueError("'apd_gain' not specified in detector characteristics.")

    @property
    def roic_gain(self) -> float:
        """Get roic gainn."""
        if self._roic_gain:
            return self._roic_gain
        else:
            raise ValueError("'roic_gain' not specified in detector characteristics.")

    @property
    def node_capacitance(self) -> float:
        """Get node capacitance."""
        self._node_capacitance = self.bias_to_node_capacitance_saphira(
            self.avalanche_bias
        )
        return self._node_capacitance

    @property
    def charge_to_volt_conversion(self):
        """Get charge to voltage conversion factor."""
        self._charge_to_volt_conversion = self.detector_gain_saphira(
            capacitance=self.node_capacitance, roic_gain=self.roic_gain
        )
        return self._charge_to_volt_conversion

    @property
    def adc_bit_resolution(self) -> int:
        """Get bit resolution of the Analog-Digital Converter."""
        if self._adc_bit_resolution:
            return self._adc_bit_resolution
        else:
            raise ValueError(
                "'adc_bit_resolution' not specified in detector characteristics."
            )

    @adc_bit_resolution.setter
    def adc_bit_resolution(self, value: int) -> None:
        """Set bit resolution of the Analog-Digital Converter."""
        self._adc_bit_resolution = value

    @property
    def adc_voltage_range(self) -> tuple[float, float]:
        """Get voltage range of the Analog-Digital Converter."""
        if self._adc_voltage_range:
            return self._adc_voltage_range
        else:
            raise ValueError(
                "'adc_voltage_range' not specified in detector characteristics."
            )

    @adc_voltage_range.setter
    def adc_voltage_range(self, value: tuple[float, float]) -> None:
        """Set voltage range of the Analog-Digital Converter."""
        self._adc_voltage_range = value

    @property
    def full_well_capacity(self) -> float:
        """Get Full well capacity."""
        if self._full_well_capacity:
            return self._full_well_capacity
        else:
            raise ValueError(
                "'full_well_capacity' not specified in detector characteristics."
            )

    @full_well_capacity.setter
    def full_well_capacity(self, value: float) -> None:
        """Set Full well capacity."""
        if value not in range(10000001):
            raise ValueError("'full_well_capacity' must be between 0 and 1e+7.")

        self._full_well_capacity = value

    @property
    def system_gain(self) -> float:
        """Get system gain."""
        return (
            self.quantum_efficiency
            * self.avalanche_gain
            * self.charge_to_volt_conversion
            * 2**self.adc_bit_resolution
        ) / (max(self.adc_voltage_range) - min(self.adc_voltage_range))

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes

    @staticmethod
    def bias_to_node_capacitance_saphira(bias: float) -> float:
        """Pixel integrating node capacitance in F.

        The below interpolates empirical published data, however note that
        Node C = Charge Gain / Voltage Gain
        So can be calculated by measuring V gain (varying PRV) and chg gain (PTC); see [2]

        Parameters
        ----------
        bias: float

        Returns
        -------
        output_capacitance: float
        """
        if bias < 1:
            raise ValueError(
                "Warning! Node capacitance calculation is inaccurate for bias voltages <1 V!"
            )

        # From [2] (Mk13 ME1000; data supplied by Leonardo):
        bias_list = [1, 1.5, 2.5, 3.5, 4.5, 6.5, 8.5, 10.5]
        capacitance = [46.5, 41.3, 37.3, 34.8, 33.2, 31.4, 30.7, 30.4]

        output_capacitance = float(np.interp(x=bias, xp=bias_list, fp=capacitance))

        return output_capacitance * 1.0e-15

    @staticmethod
    def bias_to_gain_saphira(bias: float) -> float:
        """Calculate gain from bias.

        The formula ignores the soft knee between the linear and unity gain ranges,
        but should be close enough. [2] (Mk13 ME1000)

        Parameters
        ----------
        bias: float

        Returns
        -------
        gain: float
        """

        gain = 2 ** ((bias - 2.65) / 2.17)

        if gain < 1.0:
            gain = 1.0  # Unity gain is lowest

        return gain

    @staticmethod
    def gain_to_bias_saphira(gain: float) -> float:
        """Calculate bias from gain.

        The formula ignores the soft knee between the linear and
        unity gain ranges, but should be close enough. [2] (Mk13 ME1000)

        Parameters
        ----------
        gain: float

        Returns
        -------
        bias: float
        """

        bias = (2.17 * math.log2(gain)) + 2.65

        return bias

    @staticmethod
    def detector_gain_saphira(capacitance: float, roic_gain: float) -> float:
        """Saphira detector gain.

        Parameters
        ----------
        capacitance: float
        roic_gain: float

        Returns
        -------
        float
        """
        # Late import to speedup start-up time
        import astropy.constants as const

        return roic_gain * (const.e.value / capacitance)

    def to_dict(self) -> Mapping:
        """Get the attributes of this instance as a `dict`."""
        if self._avalanche_gain and self._pixel_reset_voltage:
            dct = {
                "avalanche_gain": self._avalanche_gain,
                "common_voltage": None,
                "pixel_reset_voltage": self._pixel_reset_voltage,
            }
        elif self._avalanche_gain and self._common_voltage:
            dct = {
                "avalanche_gain": self._avalanche_gain,
                "common_voltage": self._common_voltage,
                "pixel_reset_voltage": None,
            }
        elif self._common_voltage and self._pixel_reset_voltage:
            dct = {
                "avalanche_gain": None,
                "common_voltage": self._common_voltage,
                "pixel_reset_voltage": self._pixel_reset_voltage,
            }
        else:
            raise NotImplementedError

        other_dct = {
            "quantum_efficiency": self._quantum_efficiency,
            "full_well_capacity": self._full_well_capacity,
            "adc_voltage_range": self._adc_voltage_range,
            "adc_bit_resolution": self._adc_bit_resolution,
            "roic_gain": self._roic_gain,
        }

        return dct | other_dct

    @classmethod
    def from_dict(cls, dct: Mapping):
        """Create a new instance from a `dict`."""
        if "adc_voltage_range" in dct:
            new_dct: Mapping = dicttoolz.dissoc(dct, "adc_voltage_range")
            adc_voltage_range = dct["adc_voltage_range"]

            if adc_voltage_range is not None:
                adc_voltage_range = tuple(adc_voltage_range)

            return cls(adc_voltage_range=adc_voltage_range, **new_dct)

        return cls(**dct)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
"""Configuration loader."""

from collections.abc import Iterator, Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path
from shutil import copy2
from typing import IO, TYPE_CHECKING, Any, Optional, Union

import yaml

from pyxel import __version__ as version
from pyxel.detectors import (
    APD,
    CCD,
    CMOS,
    MKID,
    APDCharacteristics,
    APDGeometry,
    CCDGeometry,
    Characteristics,
    CMOSGeometry,
    Environment,
    MKIDGeometry,
)
from pyxel.exposure import Exposure, Readout
from pyxel.observation import Observation, ParameterValues
from pyxel.outputs import CalibrationOutputs, ExposureOutputs, ObservationOutputs
from pyxel.pipelines import DetectionPipeline, FitnessFunction, ModelFunction

if TYPE_CHECKING:
    from pyxel.calibration import Algorithm, Calibration


@dataclass
class Configuration:
    """Configuration class."""

    pipeline: DetectionPipeline

    # Running modes
    exposure: Optional[Exposure] = None
    observation: Optional[Observation] = None
    calibration: Optional["Calibration"] = None

    # Detectors
    ccd_detector: Optional[CCD] = None
    cmos_detector: Optional[CMOS] = None
    mkid_detector: Optional[MKID] = None
    apd_detector: Optional[APD] = None

    def __post_init__(self):
        # Sanity checks
        running_modes = [self.exposure, self.observation, self.calibration]
        num_running_modes: int = sum(el is not None for el in running_modes)

        if num_running_modes != 1:
            raise ValueError(
                "Expecting only one running mode: "
                "'exposure', 'observation' or 'calibration'."
            )

        detectors = [
            self.ccd_detector,
            self.cmos_detector,
            self.mkid_detector,
            self.apd_detector,
        ]
        num_detectors = sum(el is not None for el in detectors)

        if num_detectors != 1:
            raise ValueError(
                "Expecting only one detector: 'ccd_detector', 'cmos_detector', 'mkid_detector' or 'apd_detector'."
            )

    @property
    def running_mode(self) -> Union[Exposure, Observation, "Calibration"]:
        """Get current running mode."""
        if self.exposure is not None:
            return self.exposure
        elif self.observation is not None:
            return self.observation
        elif self.calibration is not None:
            return self.calibration
        else:
            raise NotImplementedError

    @property
    def detector(self) -> Union[CCD, CMOS, MKID, APD]:
        """Get current detector."""
        if self.ccd_detector is not None:
            return self.ccd_detector
        elif self.cmos_detector is not None:
            return self.cmos_detector
        elif self.mkid_detector is not None:
            return self.mkid_detector
        elif self.apd_detector is not None:
            return self.apd_detector
        else:
            raise NotImplementedError


def load(yaml_file: Union[str, Path]) -> Configuration:
    """Load configuration from a ``YAML`` file.

    Parameters
    ----------
    yaml_file: str or Path

    Returns
    -------
    configuration: Configuration
    """
    filename = Path(yaml_file).resolve()
    if not filename.exists():
        raise FileNotFoundError(f"Cannot find configuration file '{filename}'.")
    with filename.open("r") as file_obj:
        dct = load_yaml(file_obj)

    return _build_configuration(dct)


def loads(yaml_string: str) -> Configuration:
    """Load configuration from a ``YAML`` string.

    Parameters
    ----------
    yaml_string: str

    Returns
    -------
    configuration: Configuration
    """
    dct = load_yaml(yaml_string)
    return _build_configuration(dct)


def load_yaml(stream: Union[str, IO]) -> Any:
    """Load a ``YAML`` document.

    Parameters
    ----------
    stream

    Returns
    -------
    result: dict

    """
    result = yaml.load(stream, Loader=yaml.SafeLoader)
    return result


def to_exposure_outputs(dct: dict) -> ExposureOutputs:
    """Create a ExposureOutputs class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    ExposureOutputs
    """
    return ExposureOutputs(**dct)


def to_readout(dct: Optional[dict] = None) -> Readout:
    """Create a Readout class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    Readout
    """
    if dct is None:
        dct = {}
    return Readout(**dct)


def to_exposure(dct: dict) -> Exposure:
    """Create a Exposure class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    Single
    """
    dct.update({"outputs": to_exposure_outputs(dct["outputs"])})
    dct.update({"readout": to_readout(dct.get("readout"))})

    return Exposure(**dct)


def to_observation_outputs(dct: dict) -> ObservationOutputs:
    """Create a ObservationOutputs class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    ObservationOutputs
    """
    return ObservationOutputs(**dct)


def to_parameters(dct: dict) -> ParameterValues:
    """Create a ParameterValues class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    ParameterValues
    """
    return ParameterValues(**dct)


def to_observation(dct: dict) -> Observation:
    """Create a Parametric class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    Observation
    """
    dct.update(
        {"parameters": [to_parameters(param_dict) for param_dict in dct["parameters"]]}
    )
    dct.update({"outputs": to_observation_outputs(dct["outputs"])})
    dct.update({"readout": to_readout(dct.get("readout"))})

    return Observation(**dct)


def to_calibration_outputs(dct: dict) -> CalibrationOutputs:
    """Create a CalibrationOutputs class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    CalibrationOutputs
    """
    return CalibrationOutputs(**dct)


def to_algorithm(dct: dict) -> "Algorithm":
    """Create an Algorithm class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    Algorithm
    """
    # Late import to speedup start-up time
    from pyxel.calibration import Algorithm

    return Algorithm(**dct)


def to_fitness_function(dct: dict) -> FitnessFunction:
    """Create a callable from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    callable
    """
    func: str = dct["func"]
    arguments: Optional[Mapping[str, Any]] = dct.get("arguments")

    return FitnessFunction(func=func, arguments=arguments)


def to_calibration(dct: dict) -> "Calibration":
    """Create a Calibration class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    Calibration
    """
    # Late import to speedup start-up time
    from pyxel.calibration import Calibration

    dct.update({"outputs": to_calibration_outputs(dct["outputs"])})
    dct.update({"fitness_function": to_fitness_function(dct["fitness_function"])})
    dct.update({"algorithm": to_algorithm(dct["algorithm"])})
    dct.update(
        {"parameters": [to_parameters(param_dict) for param_dict in dct["parameters"]]}
    )
    dct["result_input_arguments"] = [
        to_parameters(value) for value in dct.get("result_input_arguments", {})
    ]
    dct.update({"readout": to_readout(dct.get("readout"))})

    return Calibration(**dct)


def to_ccd_geometry(dct: dict) -> CCDGeometry:
    """Create a CCDGeometry class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    CCDGeometry
    """

    return CCDGeometry(**dct)


def to_cmos_geometry(dct: dict) -> CMOSGeometry:
    """Create a CMOSGeometry class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    CMOSGeometry
    """
    return CMOSGeometry(**dct)


def to_mkid_geometry(dct: dict) -> MKIDGeometry:
    """Create a MKIDGeometry class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    MKIDGeometry
    """
    return MKIDGeometry(**dct)


def to_apd_geometry(dct: dict) -> APDGeometry:
    """Create a APDGeometry class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    APDGeometry
    """
    return APDGeometry(**dct)


def to_environment(dct: Optional[dict]) -> Environment:
    """Create an Environment class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    Environment
    """
    if dct is None:
        dct = {}
    return Environment(**dct)


def to_ccd_characteristics(dct: Optional[dict]) -> Characteristics:
    """Create a CCDCharacteristics class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    CCDCharacteristics
    """
    if dct is None:
        dct = {}
    return Characteristics(**dct)


def to_cmos_characteristics(dct: Optional[dict]) -> Characteristics:
    """Create a CMOSCharacteristics class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    CMOSCharacteristics
    """
    if dct is None:
        dct = {}
    return Characteristics(**dct)


def to_mkid_characteristics(dct: Optional[dict]) -> Characteristics:
    """Create a MKIDCharacteristics class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    MKIDCharacteristics
    """
    if dct is None:
        dct = {}
    return Characteristics(**dct)


def to_apd_characteristics(dct: Optional[dict]) -> APDCharacteristics:
    """Create a APDCharacteristics class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    APDCharacteristics
    """
    if dct is None:
        dct = {}
    return APDCharacteristics(**dct)


def to_ccd(dct: dict) -> CCD:
    """Create a CCD class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    CCD
    """
    return CCD(
        geometry=to_ccd_geometry(dct["geometry"]),
        environment=to_environment(dct["environment"]),
        characteristics=to_ccd_characteristics(dct["characteristics"]),
    )


def to_cmos(dct: dict) -> CMOS:
    """Create a :term:`CMOS` class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    CMOS
    """
    return CMOS(
        geometry=to_cmos_geometry(dct["geometry"]),
        environment=to_environment(dct["environment"]),
        characteristics=to_cmos_characteristics(dct["characteristics"]),
    )


def to_mkid_array(dct: dict) -> MKID:
    """Create an MKIDarray class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    MKID-array
    """
    return MKID(
        geometry=to_mkid_geometry(dct["geometry"]),
        environment=to_environment(dct["environment"]),
        characteristics=to_mkid_characteristics(dct["characteristics"]),
    )


def to_apd(dct: dict) -> APD:
    """Create an APDarray class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    MKID-array
    """
    return APD(
        geometry=to_apd_geometry(dct["geometry"]),
        environment=to_environment(dct["environment"]),
        characteristics=to_apd_characteristics(dct["characteristics"]),
    )


def to_model_function(dct: dict) -> ModelFunction:
    """Create a ModelFunction class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    ModelFunction
    """
    return ModelFunction(**dct)


def to_pipeline(dct: dict) -> DetectionPipeline:
    """Create a DetectionPipeline class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    DetectionPipeline
    """
    for model_group_name in dct.keys():
        models_list: Optional[Sequence[dict]] = dct[model_group_name]

        if models_list is None:
            models: Optional[Sequence[ModelFunction]] = None
        else:
            models = [to_model_function(model_dict) for model_dict in models_list]

        dct.update({model_group_name: models})
    return DetectionPipeline(**dct)


def _build_configuration(dct: dict) -> Configuration:
    """Create a Configuration class from a dictionary.

    Parameters
    ----------
    dct

    Returns
    -------
    Configuration
    """
    pipeline: DetectionPipeline = to_pipeline(dct["pipeline"])

    # Sanity checks
    keys_running_mode: Sequence[str] = [
        "exposure",
        "observation",
        "calibration",
    ]
    num_running_modes: int = sum(key in dct for key in keys_running_mode)
    if num_running_modes != 1:
        keys = ", ".join(map(repr, keys_running_mode))
        raise ValueError(f"Expecting only one running mode: {keys}")

    keys_detectors: Sequence[str] = [
        "ccd_detector",
        "cmos_detector",
        "mkid_detector",
        "apd_detector",
    ]
    num_detector: int = sum(key in dct for key in keys_detectors)
    if num_detector != 1:
        keys = ", ".join(map(repr, keys_detectors))
        raise ValueError(f"Expecting only one detector: {keys}")

    running_mode: dict[str, Union[Exposure, Observation, "Calibration"]] = {}
    if "exposure" in dct:
        running_mode["exposure"] = to_exposure(dct["exposure"])
    elif "observation" in dct:
        running_mode["observation"] = to_observation(dct["observation"])
    elif "calibration" in dct:
        running_mode["calibration"] = to_calibration(dct["calibration"])
    else:
        raise ValueError("No mode configuration provided.")

    detector: dict[str, Union[CCD, CMOS, MKID, APD]] = {}
    if "ccd_detector" in dct:
        detector["ccd_detector"] = to_ccd(dct["ccd_detector"])
    elif "cmos_detector" in dct:
        detector["cmos_detector"] = to_cmos(dct["cmos_detector"])
    elif "mkid_detector" in dct:
        detector["mkid_detector"] = to_mkid_array(dct["mkid_detector"])
    elif "apd_detector" in dct:
        detector["apd_detector"] = to_apd(dct["apd_detector"])
    else:
        raise ValueError("No detector configuration provided.")

    configuration: Configuration = Configuration(
        pipeline=pipeline,
        **running_mode,  # type: ignore
        **detector,  # type: ignore
    )

    return configuration


def save(input_filename: Union[str, Path], output_dir: Path) -> Path:
    """Save a copy of the input ``YAML`` file to output directory.

    Parameters
    ----------
    input_filename: str or Path
    output_dir: Path

    Returns
    -------
    Path
    """

    input_file = Path(input_filename)
    copy2(input_file, output_dir)

    # TODO: sort filenames ?
    pattern: str = f"*{input_file.suffix}"
    copied_input_file_it: Iterator[Path] = output_dir.glob(pattern)
    copied_input_file: Path = next(copied_input_file_it)

    with copied_input_file.open("a") as file:
        file.write("\n#########")
        file.write(f"\n# Pyxel version: {version}")
        file.write("\n#########")

    return copied_input_file



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


"""Functions to add new models."""

import logging
import sys
import time
from pathlib import Path


def create_model(newmodel: str) -> None:
    """Create a new module using pyxel/templates/MODELTEMPLATE.py.

    Parameters
    ----------
    newmodel : str
        modeltype/modelname
    """

    location, model_name = get_name_and_location(newmodel)

    # Is not working on UNIX AND Windows if I do not use os.path.abspath
    template_str = "_TEMPLATE"
    template_location = "_LOCATION"

    # Copying the template with the user defined model_name instead
    import pyxel

    folder: Path = Path(pyxel.__file__).parent
    src: Path = folder.joinpath("templates").resolve()
    dest: Path = folder.joinpath(f"models/{location}").resolve()

    if not src.exists():
        raise FileNotFoundError(f"Folder '{src}' does not exists !")

    try:
        # Replacing all templates in filenames and directories by model_name
        filename: Path
        for filename in src.glob("**/*"):
            if filename.suffix == ".pyc" or filename.name == "__pycache__":
                continue

            relative_filename: str = str(filename.relative_to(src))
            new_relative_filename = relative_filename.replace(template_str, model_name)

            # Get a destination filename
            new_filename: Path = dest.joinpath(new_relative_filename)

            # Create new modified content
            content: str = filename.read_text()
            new_content: str = (
                content.replace(template_str, model_name)
                .replace(template_location, location)
                .replace("%(date)", time.ctime())
            )

            # Save this content in the destination filename
            new_filename.parent.mkdir(parents=True, exist_ok=True)
            new_filename.write_text(new_content)

        logging.info("Module %r created.", model_name)
        print(f"Module {model_name!r} created in {dest!r}.")

    except FileExistsError:
        logging.info("%r already exists, folder not created", dest)
        raise
    except OSError as exc:
        # Any error saying that the directory doesn't exist
        logging.critical("%r not created. Error: %s", model_name, exc)
        raise


def get_name_and_location(newmodel: str) -> tuple[str, str]:
    """Get name and location of new model from string modeltype/modelname.

    Parameters
    ----------
    newmodel: str

    Returns
    -------
    location: str
    model_name: str
    """

    try:
        location, model_name = newmodel.split("/")
    except Exception:
        sys.exit(
            f"""
        Can't create model {newmodel!r}, please use location/newmodelname
        as an argument for creating a model
        """
        )
    return location, model_name
